/*
 *  squelet.cc
 *  Anet
 *
 *  Created by Nicolas on Wed Apr 14 2004.
 *  Copyright (c) 2004 IEF. All rights reserved.
 *
 */

#include "anet++.h"
#include <stream.h>

void SkeletConnex(PvarInt& iSource, PvarInt& iDest);
void SkeletBasic(PvarInt& iSource, PvarInt& iDest);

int main(int argc, char **argv)
{
	unsigned    width, height, nbthread;

    if (argc < 5) {
        cerr << "ANET> Wrong parameters" << endl;
        cerr << "ANET> usage : " << argv[0] << " src nbthread F/O B/C [dest]" << endl;
		cerr << "ANET>         src = image binaire avec fond noir et objet blanc !" << endl;
		cerr << "ANET>         dest = nom de fichier de sortie (pgm binaire)" << endl;
		cerr << "ANET>         F = avec fond rajoute, O = sans fond rajoute" << endl;
		cerr << "ANET>         B = centres de cercles maximums, C = squelette connexe" << endl;
        return 1;
    }
	
	nbthread = atoi(argv[2]);
	if (nbthread < 1) nbthread = 1;
	
	if (Anet::ImageFileSize(argv[1], &width, &height)) {
        fprintf(stderr, "Erreur d'ouverture de fichier image\n");
        exit(1);
    }
	
    if (Anet::Initialize(width, height, nbthread)) {
        cerr << "ANET> Initialization error, program exiting" << endl;
        exit(1);
	}

	PvarInt iSource, iDest(255);
	
	// fond noir (0) et objet blanc (>0)
	iSource.ReadFile(argv[1]);
	cout << iSource.Width() << "x" << iSource.Height() << endl;

	NEW_TIME(skelet);
	START_TIME(skelet);
	if (argv[4][0] == 'B')
		SkeletBasic(iSource, iDest);
	else
		SkeletConnex(iSource, iDest);
	END_TIME(skelet);
	PRINT_TIME(skelet, "Skeletonization");
	
	if (argv[3][0] == 'F') {
		// ajoute le fond sur l'image resultat
		WHERE(iSource == 0)
			iDest = 0;
		ENDWHERE;
	}
	if (argc > 5)
		iDest.WriteFile(argv[5], "PGM5");

    Anet::End();
	
    return 0;
}

#define NR_ALLOWED

#ifndef NR_ALLOWED
// calcul du "crossing number" dans iCN à partir des distances iDist
void CrossingNumber(PvarInt& iDist, PvarInt& iCN)
{
	PvarInt iNeighbA, iNeighbB;
	Subnet  mgNeighb;
	
	iCN = 0;
	
	for (unsigned i=0; i < ANET_MASK_SIZE; i++) {
		unsigned j = (i+1) % ANET_MASK_SIZE;
		// prepare graph
		mgNeighb = (1 << i);
		///mgNeighb.Mask(Torus);
		// valeur du premier voisin
		iNeighbA.PlusStepAssocEx(mgNeighb, iDist);

		mgNeighb = (1 << j);
		//mgNeighb.Mask(Torus);
		// valeur du second voisin
		iNeighbB.PlusStepAssocEx(mgNeighb, iDist);
		
		// on ajoute au "crossing number" la valeur
		// absolue de la difference entre les deux voisins
		iNeighbA -= iNeighbB;
		iNeighbA.SelfAbs();
		iCN += iNeighbA;
	}
}

#else

BEGIN_NR_OPERATOR(CrossingNumber, Int)
	unsigned i;
	VVALUE = 0;
	for (i = 0; i < VNMAX; i++) {
		if ( VMASK(i) && VMASK((i+1)%VNMAX) )
			VVALUE += abs(PVALUE((i+1)%VNMAX) - PVALUE(i));
	}
END_NR_OPERATOR

void CrossingNumber(PvarInt& iDist, PvarInt& iCN)
{
	Subnet  mgMesh(0xff);
	
	iCN.StepAssocNr(NROP(CrossingNumber,Int), mgMesh, iDist);
}

#endif

#ifndef NR_ALLOWED
// calcul du nombre de connectivité dans iCN à partir des distances iDist
void ConnectNumber(PvarInt& iDist, PvarInt& iCN)
{
	PvarInt iNeighb0, iNeighb1, iNeighb2;
	Subnet  mgNeighb;
	
	iCN = 0;
	
	for (unsigned i = 0; i < ANET_MASK_SIZE; i+=2) {
		unsigned j = i+1, k = (i+2)%ANET_MASK_SIZE;
		//prepare graph
		mgNeighb = (1 << i);
		iNeighb0.PlusStepAssocEx(mgNeighb, iDist);

		mgNeighb = (1 << j);
		iNeighb1.PlusStepAssocEx(mgNeighb, iDist);

		iNeighb0.MaxEqual(iNeighb1);

		mgNeighb = (1 << k);
		iNeighb1.PlusStepAssocEx(mgNeighb, iDist);
		
		iNeighb0 -= iNeighb1;

		iCN += iNeighb0;
	}
	iCN -= 1;
}
#else

BEGIN_NR_OPERATOR(ConnectNumber, Int)
#define MAX(a,b) (((a)>(b))? (a) : (b))
	unsigned i;
	VVALUE = 0;
	for (i = 0; i < VNMAX; i+=2) {
		VVALUE += (MAX(PVALUE((i+1)%VNMAX), PVALUE((i+2)%VNMAX)) - PVALUE(i));
	}
	VVALUE -= 1;
#undef MAX
END_NR_OPERATOR

void ConnectNumber(PvarInt& iDist, PvarInt& iCN)
{
	Subnet  mgMesh(0xff);
	
	iCN.StepAssocNr(NROP(ConnectNumber,Int), mgMesh, iDist);
}

#endif

// calcul des "saddle points" aux coins de meme valeur
void SaddleEquals(PvarInt& iDist, PvarInt& iSaddle)
{
	Subnet		mgEquals(0x00), mgMasked;
	ANET_MASK   corners[4] = {(mE|mNE|mN), (mW|mNW|mN), (mE|mSE|mS), (mW|mSW|mS)};

	LinkWithEquals(mgEquals, iDist, iDist);
	for (unsigned c = 0; c < 4; c++) {
		mgMasked = mgEquals;
		mgMasked &= corners[c];
		WHERE(mgMasked == corners[c])
			iSaddle = iDist;
		ENDWHERE;
	}
}

void PrewittGradient(PvarInt& iDist, Subnet& mgGradient)
{
	PvarInt		pGrad, nGrad, vGrad, pGradAbs;
	Subnet		mgGrad, mgMesh(0xff);
	
	ANET_MASK   nMask[4] = {(mNW|mN|mNE), (mN|mNE|mE), (mNE|mE|mSE), (mW|mNW|mN)};
	ANET_MASK   pMask[4] = {(mSE|mS|mSW), (mS|mSW|mW), (mSW|mW|mNW), (mE|mSE|mS)};
	ANET_MASK   rpMask[4] = {mN, mNE, mE, mNW};
	ANET_MASK   rnMask[4] = {mS, mSW, mW, mSE};
	
	vGrad = 0;
	mgGradient = 0x00;
	//mgMesh.Mask(Torus);
	
	for (unsigned m = 0; m < 4; m++) {
		mgGrad = pMask[m];
		pGrad.PlusStepAssocEx(mgGrad, iDist);
		mgGrad = nMask[m];
		nGrad.PlusStepAssocEx(mgGrad, iDist);
		
		pGrad -= nGrad; // gradient values in pGrad
		pGradAbs = pGrad;
		pGradAbs.SelfAbs();
		
		WHERE(pGradAbs == vGrad)
			WHERE(pGrad < 0)
				mgGradient |= rnMask[m];
			ELSEWHERE
				mgGradient |= rpMask[m];
			ENDWHERE;
		ENDWHERE;
		WHERE(pGradAbs > vGrad)
			WHERE(pGrad < 0)
				mgGradient = rnMask[m];
			ELSEWHERE
				mgGradient = rpMask[m];
			ENDWHERE;
			vGrad = pGradAbs;
		ENDWHERE;
	}
}

void Reduction(PvarInt& iSkelet)
{
	PvarInt		iMarks(0);
	PvarInt		iM, iM2; //, iFIntern;
	Subnet		mG; //(0xff);
	PvarInt		bMarks(0);

	WHERE(iSkelet != 0) { iMarks = 1; } ENDWHERE;
	//iFIntern.PlusStepAssocEx(mG, iMarks);
	for (unsigned i = 0; i < ANET_MASK_SIZE; i++) {
		unsigned j = (i+2) % ANET_MASK_SIZE;
		unsigned k = (i+5) % ANET_MASK_SIZE;
		
		mG = (1<<i) | (1<<j); //mG.Mask(Torus);
		iM.PlusStepAssoc(mG, iMarks);

		mG = (1<<k); //mG.Mask(Torus);
		iM2.PlusStepAssoc(mG, iMarks);

		
		WHERE((iM == 3) && (iM2 == 1))
			iMarks = 0;
		ENDWHERE;
	}

	WHERE(iMarks == 0)
		iSkelet = 0;
	ENDWHERE;
}

#define SADDLE
#define LINKS
#define THIN
// iSource = 255 (blanc) dans les objets, O (noir) dans le fond !!
void SkeletConnex(PvarInt& iSource, PvarInt& iDest)
{
	Subnet  mgSaddle(0x00), mgLinks(0x00);
	PvarInt iCN;
	Subnet  mgMesh=0xff;
	PvarInt iMax, iDist = 0, iSkelet = 0;
	Subnet mgLinks2=0;

	/**************** Distance map ****************/
	//initialisations
	mgMesh.Mask(Torus);
	WHERE(iSource > 0)
		iDist = 0xffffff;
	ENDWHERE;
	iDist.MincDirectAssoc(mgMesh, iDist);
	// calcul des centres de cercles maximums
	iMax.MaxStepAssoc(mgMesh, iDist);
	WHERE(iMax == iDist)
		iSkelet = iDist;
	ENDWHERE;

#ifdef SADDLE
	/**************** Saddle points ****************/
	// calcul des "saddle points" aux coins de meme valeur
	SaddleEquals(iDist, iSkelet);
	// calcul des "crossing number"
	NEW_TIME(saddle_lower);
	START_TIME(saddle_lower);
	CrossingNumber(iDist, iCN);
	END_TIME(saddle_lower);
	PRINT_TIME(saddle_lower, "Crossing number");
	// deux composantes 4-connexes si iCN >=6
	WHERE(iCN >= 6)
		iSkelet = iDist;
	ENDWHERE;

	// calcul des "connectivity numbers"
	NEW_TIME(saddle_greater);
	START_TIME(saddle_greater);
	ConnectNumber(iDist, iCN);
	END_TIME(saddle_greater);
	PRINT_TIME(saddle_greater, "Connectivity number");
	// iCN = nombre de composantes 8-connexes
	WHERE(iCN >= 2)
		iSkelet = iDist;
	ENDWHERE;
#endif /*SADDLE*/
	
#ifdef LINKS
	/**************** Links points ****************/
	//PvarInt iAddress, iMaxAddress;	
	//// add links points : follow ascendant gradient
	LinkWithSups(mgLinks, iDist, iDist);

	// OU AVEC GRADIENT DE PREWITT
	PrewittGradient(iDist, mgLinks2);
	// on inverse le gradient pour diffuser le max sur le squelette
	mgLinks.Invert();
	mgLinks &= mgLinks2;
	iSkelet.MaxDirectAssoc(mgLinks, iSkelet);
	//END_TIME(links);
	//PRINT_TIME(links, "Links Points");
#endif /*LINKS*/

#ifdef THIN
	Reduction(iSkelet);
#endif /*THIN*/
	// partout ou iSkelet est non nul les pixels sont noirs
	WHERE(iSkelet != 0)
		iDest = 0;
	ENDWHERE;
}

// iSource = 255 (blanc) dans les objets, O (noir) dans le fond !!
void SkeletBasic(PvarInt& iSource, PvarInt& iDest)
{
	Subnet mgMesh = 0xff;
	PvarInt iDist(0), iMax;
	
	mgMesh.Mask(Torus);
	
	iDest = 255;
	WHERE(iSource > 0){
		iDist = 0xffff;
	}ENDWHERE;
	
	iDist.MincDirectAssoc(mgMesh,iDist);
	iMax.MaxStepAssoc(mgMesh,iDist);
	WHERE((iSource > 0) && (iDist == iMax)){
			iDest = 0;
	}ENDWHERE;

}
