/*
 *  voroseg-mm.cc
 *  Anet
 *
 *  segmentation de Vorono
 *
 *  Created by Nicolas on Sun Apr 18 2004.
 *  Copyright (c) 2004 IEF. All rights reserved.
 *
 */

#include "anet++.h"
#include "vnet++.h"
#include <stream.h>
#include <time.h>

/*
 Calcul de l'homogeneite en  utilisant max-min
 Renvoie dans iRes la difference max-min de chaque region.
 Il y a un seul resultat par region, duplique sur tous les pixels de
 la region.
 */
void  Homogeneite(PvarInt &iRes, Subnet &mgRegions, PvarInt &iSource)
{
	PvarInt iMin;
	iMin.MinDirectAssoc(mgRegions, iSource);
	iRes.MaxDirectAssoc(mgRegions, iSource);
	iRes -= iMin;
}


void VoronoiRegions(PvarInt& iRacines, Subnet& mgMesh, PvarInt& iDist, Subnet& mgRegions)
{
    PvarInt iRacAddress(0), iRegions;
    Subnet mgSups;
	
	// calcul de la transformée en distances
    iDist.MincDirectAssoc(mgMesh, iRacines);
	
	// on relie les pixels à leur(s) graine(s) 
    LinkWithSups(mgSups, iDist, iDist);
	mgSups.Invert();
	mgSups.Mask(Torus);
	
	// on diffuse l'indice de la graine
	// (ou de celle qui a le plus grand indice)
	WHERE(iRacines == 0) {
		iRacAddress.SelfAddress(ANET_WITHCONTEXT);
	} ENDWHERE;
    iRegions.MaxDirectAssoc(mgSups,iRacAddress);
	
	// on relie les pixels attaches a la meme racine
    LinkWithEquals(mgRegions, iRegions, iRegions);
    mgRegions.Mask(Torus);	
}

/*
 Fabrique au plus max racine dans l'image iRacine : une racine vaut 0
 le reste est a l'infini (ie tres grand nombre)
 */
void FabriqueRacines(PvarInt &iRacines, int max)
{
    int r, i=0, size=iRacines.Size();
    //srandom((int)time(NULL));
    srand(0);  // pour obtenir toujours la meme sequence pseudo-aleatoire
    iRacines = size;  //tous les pixels prennent la valeur size
					  //on designe aleatoirement au plus max racines
	do {
		r = random() % size;
        if(iRacines[r]!=0){iRacines.Set(r,0);i++;}
	} while (i < max);
}

void AjoutRacines(Subnet& mgRegions, PvarInt& iRacines,
				  PvarInt& iDist, PvarInt& iSource,
				  PvarInt& iNouvRac)
{
	PvarInt iLabels, iLabelMax;
	
	// calcul des distances max
	iLabelMax.MaxDirectAssoc(mgRegions, iDist);
	// on choisit les pixels pres du bord
	iLabelMax *= 3; iLabelMax /= 4;
	
	// on melange les adresses au hazard
	iLabels.SelfAddress();
	iLabels ^= random();
	
	WHERE(iDist != iLabelMax || iRacines == 0) {
		iLabels = -1;
	} ENDWHERE;
	iLabelMax.MaxDirectAssoc(mgRegions, iLabels);
	
	iNouvRac = iRacines;
	WHERE(iLabels == iLabelMax) {
		iNouvRac = 0;
	} ENDWHERE;
	
	/* combien de racines ? */
	/*	PvarInt iUn(1);
	WHERE(iNouvRac == 0) {
		nbrac = iUn.GlobalPlus(ANET_WITHCONTEXT) - nbrac;
		cout << nbrac << " nouvelles racines" << endl;
	} ENDWHERE;*/
}

int Split(PvarInt& iSource, int hom, int nbrinit, Subnet& mgRegions, PvarInt& iRacines) 
{
	PvarInt iFin, iHom, iTest, iDist;
	PvarInt iNouvRac;
	Subnet  mgMesh(0xff);
	
	mgMesh.Mask(Torus);
	iFin = 1;
	//cout << "Debut du split\n" ;
	unsigned int iter = 0, hPixels ;
	
    FabriqueRacines(iRacines, nbrinit);
	do{
		iter ++ ;
		//Calcul des regions de voronoi
		VoronoiRegions(iRacines, mgMesh, iDist, mgRegions);
		
		//Calcule l'homogeneite de chaque region.
		Homogeneite(iHom, mgRegions, iSource);
		
		//On prepare d'eventuelles nouvelles racines dans les regions de mgRegion.
		AjoutRacines(mgRegions, iRacines, iDist, iSource, iNouvRac);
		
		//On ajoute les nouvelles racines dans les regions non-homogenes"
		//iFin = 1;
		WHERE(iHom > hom){
			hPixels = iFin.GlobalPlus(ANET_WITHCONTEXT);
			//iFin = 0;
			iRacines = iNouvRac;
		}ENDWHERE;
		
		//hPixels = iFin.GlobalPlus();
		//cout << "pixels dans les regions homogenes : " << hPixels << endl;
	} while (hPixels > 0);
	//for (unsigned i = 0; i < 20; i++) cout << iHom[i] << " ";
	//cout << endl;
	//  while(hPixels < (IFin.Size()) /* - (IFin.Size()/50))*/);
	return iter;
}

int Merge(PvarInt& iSource, int hom, int smallsurf, int bigfactor, int homaver,
		  Subnet& mgRegions, Subnet& mgMerge, unsigned *nbreg)
{
	Subnet  mgMesh(0xff);
	PvarInt iMerge, iVar, iUn(1);
	mgMesh.Mask(Torus);
	int n, n2, iter=0, nbfus;
	enum{BEST, GLOBAL, LOCAL, SMALL};
	int step = BEST;
	
	Vnet	vn(mgRegions, mgMesh, VNET_SINGLETONS);
	//cout << vn.Size() << " sommets virtuels et " << vn.LinkCount() << " liens" << endl;
	
	do {
		WITH(vn) {
			VPvarInt	ivAdd, ivMaxAdd, ivMaxMin, ivCount(1);
			VPvarInt	ivMax, ivMin, ivMax2, ivMin2, ivBestMaxMin(0xffff), ivBestFus(-1);
			VSubnet		gvComp(0xff), gvMax, gvFus(0x00);
			// vpvar d'adresses
			ivAdd.SelfAddress();
			ivAdd ^= random();
			/* Calcul du max et du min de chaque region */
			ivMax.MaxReduce(iSource);
			ivMin.MinReduce(iSource);
			
			n = -1;
			if (step <= LOCAL) {
				do {
					// on cherche un voisin de plus grand indice
					ivMaxAdd.MaxStepAssocEx(gvComp, ivAdd);
					// on cree un arc depuis ce voisin dans gvMax
					LinkWithEquals(gvMax, ivMaxAdd, ivAdd);
					
					// on calcul les min et max du voisin ...
					ivMax2.MaxStepAssocEx(gvMax, ivMax);
					ivMin2.MinStepAssocEx(gvMax, ivMin);
					// ... et on ne garde que le max du max et le min du min
					VN_WHERE(ivMax > ivMax2);
						ivMax2 = ivMax;
					VN_ENDWHERE;
					VN_WHERE(ivMin < ivMin2);
						ivMin2 = ivMin;
					VN_ENDWHERE;
					// on calcul la difference max-min apres fusion
					ivMaxMin = ivMax2; ivMaxMin -= ivMin2;
					// on ne garde que la meilleure difference dans iBestMaxMin
					VN_WHERE((ivMaxMin < ivBestMaxMin) && (ivMaxAdd >= 0));
						LinkWithEquals(gvMax, ivMaxAdd, ivAdd);
						ivBestMaxMin = ivMaxMin;
						ivBestFus = ivMaxAdd;	// on garde l'indice du "meilleur" voisin
					VN_ENDWHERE;
					// on retire de gvComp les arcs des voisins explores
					gvMax.SelfComplement(); gvComp &= gvMax;
					// on s'arrete quand c'est stable
					ivMaxAdd -= ivAdd;
					n2 = n;
					n = ivMaxAdd.GlobalPlus();
				} while(n != n2);
				
				VPvarInt	ivNBestFus;
				VN_WHERE(ivBestMaxMin <= hom);
				LinkWithEquals(gvFus, ivBestFus, ivAdd);
				VN_ENDWHERE;
				//printf("%d composantes pretes pour une fusion locale\n",nbfus);
				
				if (step == BEST) {
					
					ivNBestFus.PlusStepAssocEx(gvFus, ivBestFus);
					/* on ne garde que les composantes pouvant fusionner */
					VN_WHERE((ivBestMaxMin <= hom) && (ivNBestFus == ivAdd))
						LinkWithEquals(gvFus, ivBestFus, ivAdd);
					nbfus = ivCount.GlobalPlus();
					VN_ENDWHERE;
					//printf("%d composantes vont fusionner (BEST)\n", nbfus);
					if (nbfus == 0) step+=2; // GLOBAL seems to be useless
					
				} else {
					if (step == GLOBAL) {
						
						// on calcule l'homogeneite globale des composantes connexes de gvFus
						VPvarInt	ivCCMax, ivCCMin;
						ivCCMax.MaxDirectAssoc(gvFus, ivMax);
						ivCCMin.MinDirectAssoc(gvFus, ivMin);
						
						ivMaxMin = ivCCMax - ivCCMin; // max-min dans les CC

						ivCount = 0;
						VN_WHERE((ivBestMaxMin <= hom) && ((ivMaxMin <= hom) ));
							LinkWithEquals(gvFus, ivBestFus, ivAdd);
							nbfus = ivCount.GlobalPlus();
						VN_ENDWHERE;
						//printf("%d composantes vont fusionner (GLOBAL)\n",nbfus);
						if (nbfus == 0) step++;
						
					} else { // LOCAL
						ivMaxAdd.MaxDirectAssoc(gvFus, ivAdd);
						VN_WHERE((ivBestMaxMin <= hom) && (ivAdd == ivMaxAdd))
							LinkWithEquals(gvFus, ivBestFus, ivAdd);
						nbfus = ivCount.GlobalPlus();
						VN_ENDWHERE;
						//printf("%d composantes vont fusionner (LOCAL)\n",nbfus);
						if (nbfus == 0) step++;
						//printf("step = %d\n", step);
					}
				}
			}  else if (step == SMALL) { // absorption des petites composantes
				VPvarInt	ivSmallSurface, ivAverage, ivAverageDiff, ivBestAverageDiff(0xffff);
				VPvarInt	ivSurface, ivSum;
				// surfaces des regions
				ivSurface.PlusReduce(iUn);
				// somme des pixels dans les regions
				ivSum.PlusReduce(iSource);
				
				//printf("Absorption des petites regions\n");
				ivSmallSurface = ivSurface; ivSmallSurface *= bigfactor;
				gvComp = 0x00;
				VN_WHERE(ivSurface > (smallsurf*bigfactor))
					/* On cherche les petites regions voisines*/
					LinkWithInfs(gvComp, ivSurface, ivSmallSurface);
				nbfus = ivCount.GlobalPlus();
				VN_ENDWHERE;
				//printf("%d grandes regions\n", nbfus);
				ivAverage = ivSum; ivAverage /= ivSurface;
				
				n = -1;
				ivBestFus = -1;
				do {
					// on cherche une voisine et on la relie dans gvMax
					ivMaxAdd.MaxStepAssocEx(gvComp, ivAdd);
					LinkWithEquals(gvMax, ivMaxAdd, ivAdd);
					// on calcule la valeur absolue de la difference des moyennes
					ivAverageDiff.MaxStepAssocEx(gvMax,ivAverage);
					ivAverageDiff -= ivAverage;
					ivAverageDiff.SelfAbs();
					// on ne garde que les meilleures differences
					gvMax = 0x00;
					VN_WHERE((ivSurface > (smallsurf*bigfactor)) && (ivAverage > 0) && (ivAverageDiff < ivBestAverageDiff))
						LinkWithEquals(gvMax, ivMaxAdd, ivAdd);
					ivBestAverageDiff = ivAverageDiff;
					ivBestFus = ivMaxAdd;
					VN_ENDWHERE;
					// on retire les liens explores de gvComp
					gvMax.SelfComplement();
					gvComp &= gvMax;
					n2 = n;
					n = ivMaxAdd.GlobalPlus();
					// cout << n << " restantes" << endl;
				} while(n != n2);
				gvFus = 0x00;
				VN_WHERE((ivSurface > (smallsurf*bigfactor)) && (ivBestFus > 0) && (ivBestAverageDiff < homaver) );
				LinkWithEquals(gvFus, ivBestFus, ivAdd);
				nbfus = ivCount.GlobalPlus();
				VN_ENDWHERE;
				//printf("%d grosses regions vont absorber une petite\n",nbfus);
				if (nbfus == 0) step++;
			}
		
			// on fusionne
			vn.Virtualize(gvFus);
			iter++;
		} ENDWITH;
		//cout << vn.Size() << " nouveaux sommets virtuels et " << vn.LinkCount() << " liens" << endl;
	} while (step <= SMALL);
	
	*nbreg = vn.Size();
	cout << *nbreg << " regions" << endl;
	WITH(vn) {
		VPvarInt	ivMerge;
		
		ivMerge.SelfAddress();
		ivMerge.Extend(iMerge);
	} ENDWITH;
	
	LinkWithEquals(mgMerge, iMerge, iMerge);
	mgMerge.Mask(Torus);
	return iter;
}

int main(int argc, char **argv)
{
	unsigned nbthread, width, height, nbroot, nbreg;
	int homsplit, hommerge, itersplit=0, itermerge=0, homaver;
	char	filename[256];
	
    if (argc < 7) {
        cerr << "ANET> Wrong parameters" << endl;
        cerr << "ANET> usage : " << argv[0] << " imagesrc nbroot homsplit hommerge homaver nbthread [imagedest]" << endl;
        return 1;
    }
	
	nbthread = atoi(argv[6]);
	if (nbthread < 1) nbthread = 1;
	nbroot = atoi(argv[2]);
	homsplit = atoi(argv[3]);
	hommerge = atoi(argv[4]);
	homaver = atoi(argv[5]);
	
	if (Anet::ImageFileSize(argv[1], &width, &height)) {
        fprintf(stderr, "Erreur d'ouverture de fichier image\n");
        exit(1);
    }
	
    if (Anet::Initialize(width, height, nbthread)) {
        cerr << "ANET> Initialization error, program exiting" << endl;
        exit(1);
	}
	
	PvarInt iSource, iDest(255), iRacines;
	
	iSource.ReadFile(argv[1]);
	cout << iSource.Width() << "x" << iSource.Height() << endl;
	
    Subnet	mgRegions, mgMerge;
	
	if (homsplit > -1) {
		itersplit = Split(iSource, homsplit, nbroot, mgRegions, iRacines);
		cout << itersplit << " iterations" << endl;
		if (argc > 7) {
			sprintf(filename, "%s-spl-%ds-%d-%di.pgm", argv[7], nbroot, homsplit, itersplit);
			mgRegions.WriteFile(filename, "PGM5");
		}
	} else {
		mgRegions = 0x00;
	}
	
	if (hommerge > -1) {
		//printf("hommerge = %d; homsplit = %d; smallsurf = %d\n", hommerge, homsplit, smallsurf);
		itermerge = Merge(iSource, hommerge, 6, 8, homaver, mgRegions, mgMerge, &nbreg);
		cout << itermerge << " iterations" << endl;
		if (argc > 7) {
			if (homsplit > -1)
				sprintf(filename, "%s-mm-%ds-%d-%di-%d-%d-%dr-%di.pgm",
						argv[7], nbroot, homsplit, itersplit, hommerge, homaver, nbreg,itermerge);
			else
				sprintf(filename, "%s-mm-N-N-N-N-%d-%d-%dr-%di.pgm",
						argv[7], hommerge, homaver, nbreg, itermerge);
			mgMerge.WriteFile(filename, "PGM5");
		}
	}
    Anet::End();
	
    return 0;
	
}
