/*
 *  voroseg-var.cc
 *  Anet
 *
 *  segmentation de Vorono
 *
 *  Created by Nicolas on Sun Apr 18 2004.
 *  Copyright (c) 2004 IEF. All rights reserved.
 *
 */

#include "anet++.h"
#include "vnet++.h"
#include <stream.h>
#include <time.h>

/*
 Calcul de l'homogeneite en  utilisant la variance
 Renvoie dans iRes la variance de chaque region.
 Il y a un seul resultat par region, duplique sur tous les pixels de
 la region.
 */
void  Homogeneite(PvarInt &iVar, Subnet &mgRegions, PvarInt &iSource)
{
	PvarInt iSum, iMoy, iUn(1);

	// calcul de somme et de la moyenne
	iSum.PlusAssoc(mgRegions, iUn);
	iMoy.PlusAssoc(mgRegions, iSource);
	iMoy /= iSum;
	// calcul de la variance
	iVar = iSource;
	iVar -= iMoy;
	iVar *= iVar;
	iVar.PlusAssoc(mgRegions, iVar);
	iVar /= iSum;
}

void VoronoiRegions(PvarInt& iRacines, Subnet& mgMesh, PvarInt& iDist, Subnet& mgRegions)
{
    PvarInt iRacAddress(0), iRegions;
    Subnet mgSups;
	
	// calcul de la transformée en distances
    iDist.MincDirectAssoc(mgMesh, iRacines);
	
	// on relie les pixels à leur(s) graine(s) 
    LinkWithSups(mgSups, iDist, iDist);
	mgSups.Invert();
	mgSups.Mask(Torus);
	
	// on diffuse l'indice de la graine
	// (ou de celle qui a le plus grand indice)
	WHERE(iRacines == 0) {
		iRacAddress.SelfAddress();
	} ENDWHERE;
    iRegions.MaxDirectAssoc(mgSups,iRacAddress);
	
	// on relie les pixels attaches a la meme racine
    LinkWithEquals(mgRegions, iRegions, iRegions);
    mgRegions.Mask(Torus);	
}

/*
 Fabrique au plus max racine dans l'image iRacine : une racine vaut 0
 le reste est a l'infini (ie tres grand nombre)
 */
void FabriqueRacines(PvarInt &iRacines, int max)
{
    int r, i=0, size=iRacines.Size();
    //srandom((int)time(NULL));
    srand(0);  // pour obtenir toujours la meme sequence pseudo-aleatoire
    iRacines = size;  //tous les pixels prennent la valeur size
    //on designe aleatoirement au plus max racines
	do {
		r = random() % size;
        if(iRacines[r]!=0){iRacines.Set(r,0);i++;}
	} while (i < max);
}

void AjoutRacines(Subnet& mgRegions, PvarInt& iRacines,
				  PvarInt& iDist, PvarInt& iSource,
				  PvarInt& iNouvRac)
{
	PvarInt iLabels, iLabelMax;
	
	// calcul des distances max
	iLabelMax.MaxDirectAssoc(mgRegions, iDist);
	// on choisit les pixels pres du bord
	iLabelMax *= 3; iLabelMax /= 4;

	// on melange les adresses au hazard
	iLabels.SelfAddress();
	iLabels ^= random();

	WHERE(iDist != iLabelMax || iRacines == 0) {
		iLabels = -1;
	} ENDWHERE;
	iLabelMax.MaxDirectAssoc(mgRegions, iLabels);

	iNouvRac = iRacines;
	WHERE(iLabels == iLabelMax) {
		iNouvRac = 0;
	} ENDWHERE;
}

int Split(PvarInt& iSource, int hom, int nbrinit, Subnet& mgRegions, PvarInt& iRacines) 
{
	PvarInt iFin, iHom, iTest, iDist;
	PvarInt iNouvRac;
	Subnet  mgMesh(0xff);

	mgMesh.Mask(Torus);
	iFin = 1;
	//cout << "Debut du split\n" ;
	unsigned int iter = 0, hPixels ;
	
    FabriqueRacines(iRacines, nbrinit);
	do{
		iter ++ ;
		//Calcul des regions de voronoi
		VoronoiRegions(iRacines, mgMesh, iDist, mgRegions);

		//Calcule l'homogeneite de chaque region.
		Homogeneite(iHom, mgRegions, iSource);

		//On prepare d'eventuelles nouvelles racines dans les regions de mgRegion.
		AjoutRacines(mgRegions, iRacines, iDist, iSource, iNouvRac);

		//On ajoute les nouvelles racines dans les regions non-homogenes"
		//iFin = 1;
		WHERE(iHom > hom){
			hPixels = iFin.GlobalPlus(ANET_WITHCONTEXT);
			//iFin = 0;
			iRacines = iNouvRac;
		}ENDWHERE;

		//hPixels = iFin.GlobalPlus();
		//cout << "pixels dans les regions homogenes : " << hPixels << endl;
	} while (hPixels > 0);
	//for (unsigned i = 0; i < 20; i++) cout << iHom[i] << " ";
	//cout << endl;
//  while(hPixels < (IFin.Size()) /* - (IFin.Size()/50))*/);
	return iter;
}

int Merge(PvarInt& iSource, int hom, int smallsurf, int bigfactor, int homaver,
		  Subnet& mgRegions, Subnet& mgMerge, unsigned *nbreg)
{
	Subnet  mgMesh(0xff);
	PvarInt iMerge, iSource2, iVar, iUn(1);
	mgMesh.Mask(Torus);
	int n, n2, iter=0, nbfus=0;
	enum{BEST, GLOBAL, LOCAL, SMALL};
	int step = BEST;

	Vnet	vn(mgRegions, mgMesh, VNET_SINGLETONS);
	//cout << vn.Size() << " sommets virtuels et " << vn.LinkCount() << " liens" << endl;

	iSource2 = iSource; iSource2 *= iSource;

	/* fusions reciproques */
	//cout << "**** Fusions reciproques ****" << endl;
	do {
		WITH(vn) {
			VPvarInt	ivAdd, ivMaxAdd, ivSurface, ivSum, ivSum2, ivQ, ivR, ivCount(1) ;
			VPvarInt	ivSurfaceFus, ivSumFus, ivVarFus, ivBestVarFus(0xffff), ivBestFus;
			VSubnet		gvMax(0x00), gvFus, gvComp(0xff);
			// vpvar d'adresses
			ivAdd.SelfAddress();
			ivAdd ^= random();
			// surfaces des regions
			ivSurface.PlusReduce(iUn);
			// somme des pixels dans les regions
			ivSum.PlusReduce(iSource);
			// somme des carrés dans les regions
			ivSum2.PlusReduce(iSource2);
			
			n = -1;
			if (step <= LOCAL) {
				do {
					// on cherche sa voisine la plus "grande" dans gComp
					// et on la relie dans gvMax
					ivMaxAdd.MaxStepAssocEx(gvComp, ivAdd);
					LinkWithEquals(gvMax, ivMaxAdd, ivAdd);

					// calcul de la variance que l'on obtiendrait apres fusion virtuelle
					ivSurfaceFus.PlusStepAssoc(gvMax, ivSurface); // surfaces
					ivVarFus.PlusStepAssoc(gvMax, ivSum2); //sommes des carres
					ivVarFus /= ivSurfaceFus; // sommes des carres / surfaces

					ivSumFus.PlusStepAssoc(gvMax, ivSum); // sommes des valeurs
					ivQ = ivSumFus; ivQ /= ivSurfaceFus;
					ivR = ivSumFus; ivR %= ivSurfaceFus;
					ivR *= ivQ;
					ivR <<= 1; // eq. iv3 *= 2
					ivR /= ivSurfaceFus;
					ivQ *= ivQ;
					ivQ += ivR;

					ivVarFus -= ivQ; // sommes des carres - moyennes carrees

					gvMax = 0x00;
					// on garde la meilleure variance
					VN_WHERE((ivVarFus < ivBestVarFus) && (ivMaxAdd >= 0));
						LinkWithEquals(gvMax, ivMaxAdd, ivAdd);
						ivBestVarFus = ivVarFus;
						ivBestFus = ivMaxAdd;
					VN_ENDWHERE;
					
					gvMax.SelfComplement();
					gvComp &= gvMax;
					n2 = n;
					n = ivMaxAdd.GlobalPlus();
					// cout << n << " restantes" << endl;
				} while(n != n2);
				
				gvFus = 0x00;
				VN_WHERE(ivBestVarFus <= hom);
					LinkWithEquals(gvFus, ivBestFus, ivAdd);
				VN_ENDWHERE;
				//printf("%d composantes pretes pour une fusion locale\n",nbfus);

				if (step == BEST) {
					VPvarInt	ivNBestFus;

					ivNBestFus.PlusStepAssocEx(gvFus, ivBestFus);
					/* on ne garde que les composantes pouvant fusionner */
					gvFus = 0x00;
					VN_WHERE((ivBestVarFus <= hom) && (ivNBestFus == ivAdd))
						LinkWithEquals(gvFus, ivBestFus, ivAdd);
						nbfus = ivCount.GlobalPlus();
					VN_ENDWHERE;
					//printf("%d composantes vont fusionner (BEST)\n", nbfus);
					if (nbfus == 0) step++;

				} else {
					if (step == GLOBAL) {
						
						ivSurfaceFus.PlusAssoc(gvFus, ivSurface); // surfaces
						ivVarFus.PlusAssoc(gvFus, ivSum2); //sommes des carres
						ivVarFus /= ivSurfaceFus; // sommes des carres / surfaces
						
						ivSumFus.PlusAssoc(gvFus, ivSum); // sommes des valeurs
						ivQ = ivSumFus / ivSurfaceFus;
						ivR = ivSumFus % ivSurfaceFus;
						ivR *= ivQ;
						ivR *= 2;
						ivR /= ivSurfaceFus;
						ivQ *= ivQ;
						ivQ += ivR;
						
						ivVarFus -= ivQ; // sommes des carres - moyennes carrees
						/* on ne garde que les composantes pouvant fusionner */
						gvFus = 0x00;
						VN_WHERE((ivBestVarFus <= hom) && (ivVarFus <= hom))
							LinkWithEquals(gvFus, ivBestFus, ivAdd);
							nbfus = ivCount.GlobalPlus();
						VN_ENDWHERE;
						//printf("%d composantes vont fusionner (GLOBAL)\n",nbfus);
						if (nbfus == 0) step++;

					} else if (step == LOCAL) { // LOCAL
						ivMaxAdd.MaxDirectAssoc(gvFus, ivAdd);
						gvFus = 0x00;
						VN_WHERE((ivBestVarFus <= hom) && (ivAdd == ivMaxAdd))
							LinkWithEquals(gvFus, ivBestFus, ivAdd);
							nbfus = ivCount.GlobalPlus();
						VN_ENDWHERE;
						//printf("%d composantes vont fusionner (LOCAL)\n",nbfus);
						if (nbfus == 0) step++;
					}
				}
			}  else if (step == SMALL) { // absorption des petites composantes
				VPvarInt	ivSmallSurface(ivSurface), ivAverage, ivAverageDiff, ivBestAverageDiff(0xffff);
				//printf("Absorption des petites regions\n");
				ivSmallSurface *= bigfactor;
				gvComp = 0x00;
				VN_WHERE(ivSurface > (smallsurf*bigfactor))
					/* On cherche les petites regions voisines*/
					LinkWithInfs(gvComp, ivSurface, ivSmallSurface);
					nbfus = ivCount.GlobalPlus();
				VN_ENDWHERE;
				//printf("%d grandes regions\n", nbfus);
				ivAverage = ivSum; ivAverage /= ivSurface;
				
				n = -1;
				ivBestFus = -1;
				do {
					// on cherche une voisine et on la relie dans gvMax
					ivMaxAdd.MaxStepAssocEx(gvComp, ivAdd);
					LinkWithEquals(gvMax, ivMaxAdd, ivAdd);
					// on calcule la valeur absolue de la difference des moyennes
					ivAverageDiff.MaxStepAssocEx(gvMax,ivAverage);
					ivAverageDiff -= ivAverage;
					ivAverageDiff.SelfAbs();
					// on ne garde que les meilleures differences
					gvMax = 0x00;
					VN_WHERE((ivSurface > (smallsurf*bigfactor)) && (ivAverage > 0) && (ivAverageDiff < ivBestAverageDiff))
						LinkWithEquals(gvMax, ivMaxAdd, ivAdd);
						ivBestAverageDiff = ivAverageDiff;
						ivBestFus = ivMaxAdd;
					VN_ENDWHERE;
					// on retire les liens explores de gvComp
					gvMax.SelfComplement();
					gvComp &= gvMax;
					n2 = n;
					n = ivMaxAdd.GlobalPlus();
				} while(n != n2);
				gvFus = 0x00;
				VN_WHERE((ivSurface > (smallsurf*bigfactor)) && (ivBestFus > 0) && (ivBestAverageDiff < homaver) );
					LinkWithEquals(gvFus, ivBestFus, ivAdd);
					nbfus = ivCount.GlobalPlus();
				VN_ENDWHERE;
				//printf("%d grosses regions vont absorber une petite\n",nbfus);
				if (nbfus == 0) step++;
			}
			
			// on fusionne
			if (nbfus) vn.Virtualize(gvFus);
			iter++;
		} ENDWITH;
		//cout << vn.Size() << " nouveaux sommets virtuels et " << vn.LinkCount() << " liens" << endl;
	} while (step <= SMALL);

	*nbreg = vn.Size();
	cout << *nbreg << " regions" << endl;
	WITH(vn) {
		VPvarInt	ivMerge;

		ivMerge.SelfAddress();
		ivMerge.Extend(iMerge);
	} ENDWITH;
	
	LinkWithEquals(mgMerge, iMerge, iMerge);
	mgMerge.Mask(Torus);
	return iter;
}

int main(int argc, char **argv)
{
	unsigned nbthread, width, height, nbroot, nbreg;
	int homsplit, hommerge, itersplit=0, itermerge=0, homaver;
	char	filename[256];
	
    if (argc < 7) {
        cerr << "ANET> Wrong parameters" << endl;
        cerr << "ANET> usage : " << argv[0] << " imagesrc nbroot homsplit hommerge homaver nbthread [imagedest]" << endl;
        return 1;
    }
	
	nbthread = atoi(argv[6]);
	if (nbthread < 1) nbthread = 1;
	nbroot = atoi(argv[2]);
	homsplit = atoi(argv[3]);
	hommerge = atoi(argv[4]);
	homaver = atoi(argv[5]);
	
	if (Anet::ImageFileSize(argv[1], &width, &height)) {
        fprintf(stderr, "Erreur d'ouverture de fichier image\n");
        exit(1);
    }
	
    if (Anet::Initialize(width, height, nbthread)) {
        cerr << "ANET> Initialization error, program exiting" << endl;
        exit(1);
	}
	
	PvarInt iSource, iDest(255), iRacines;
	
	iSource.ReadFile(argv[1]);
	cout << iSource.Width() << "x" << iSource.Height() << endl;
	
    Subnet	mgRegions, mgMerge;

	if (homsplit > -1) {
		itersplit = Split(iSource, homsplit, nbroot, mgRegions, iRacines);
		cout << itersplit << " iterations" << endl;
		if (argc > 7) {
			sprintf(filename, "%s-spl-%ds-%d-%di.pgm", argv[7], nbroot, homsplit, itersplit);
			mgRegions.WriteFile(filename, "PGM5");
		}
	} else {
		mgRegions = 0x00;
	}

	if (hommerge > -1) {
		//printf("hommerge = %d; homsplit = %d; smallsurf = %d\n", hommerge, homsplit, smallsurf);
		itermerge = Merge(iSource, hommerge, 6, 8, homaver, mgRegions, mgMerge, &nbreg);
		cout << itermerge << " iterations" << endl;
		if (argc > 7) {
			if (homsplit > -1)
				sprintf(filename, "%s-vmrg-%ds-%d-%di-%d-%d-%dr-%di.pgm",
						argv[7], nbroot, homsplit, itersplit, hommerge, homaver, nbreg,itermerge);
			else
				sprintf(filename, "%s-vmrg-N-N-N-N-%d-%d-%dr-%di.pgm",
						argv[7], hommerge, homaver, nbreg, itermerge);
			mgMerge.WriteFile(filename, "PGM5");
		}
	}
    Anet::End();
	
    return 0;
}
