#!/usr/bin/perl -w
# Template Preprocessor in Perl
#
# $Id: codegen.perl,v 1.1.1.1 2003/09/09 14:41:03 nico Exp $
# $Log: codegen.perl,v $
# Revision 1.1.1.1  2003/09/09 14:41:03  nico
# Version 0.9
#
# Revision 1.3  2003/07/17 12:11:11  nico
# *** empty log message ***
#
# Revision 1.2  2003/07/07 21:53:46  nico
# Diverses mises à jour
#
#
# Allows to embedd Perl commands in a file.
# Perl commands are indicated with a `:' in column 0 of a line,
# while other lines consist of regular text.
# Perl commands will be applied to the text, that allows loop,
# conditional text generation, etc.
# The text can contain either
#   - plain perl variables eg. $foo
#   - associative perl arrays reference eg $foo{'bar'} or
#     $foo{$bar}
# Other perl varaible instantiations are ignored (arrays, etc)
# $ in the text part can be left unchanged by using the @$ or
# $@ escape sequences (that will both output a simple $).
# 
# tpp works by 1) analysing the source file and generating a
# perl programm that will interpret perl code and 'print' other
# lines 2) executing this temporary file.
# 
# Usage: tpp [options] file.txx
# preprocess and generates file.xx
#
# Options : -k keep temporary files (named file.xxTMP) Useful
#              for debugging purposes
#           -e do not echo filenames on stdout


package main ;


$printlineno = 0;		# should we keep track of line
				# numbers and file names
$keeptempfiles=0;
$echofilenames = 1 ;

$datafile="";
$datafile=$ARGV[0];
shift;

while ($_ = $ARGV[0], /^-/) {
    shift;
    last if /^--$/;
    if (/^-k/) { $keeptempfiles = 1 ; }
    if (/^-e/) { $echofilenames = 0 ; }
}



$file=$ARGV[0];

die "Usage tpp [options] file.txx" if (!defined($file)) ;

$fileout=$file;
$fileout =~ s/\.t/./ ;
# pb with DOS filenames
if($file eq $fileout) {
$fileout =~ s/\.T/./ ;
}
$filetemp=$fileout;
$filetemp =~ s/$/TMP/ ;
print "$file : $fileout\n" if ($echofilenames) ;
die "Usage tpp file.txx" if ($file eq $fileout) ;

$lineno=0;

open(FILE,$file) || die "No such file ". $file;
open(FILETEMP,">$filetemp") || die "cannot open file ". $filetemp;
select(FILETEMP) ;

sub getline {
# read a line from current input file, manage line numbers,
# continuation  lines, skip perl comments.
    local ($f)=@_;
    local($l);

    $l = "" ;

  lines: {
    $_ = <$f>;
    $lineno++;

    redo lines if  (/^:[ \t]*#/)  ; # skip comments

#     if (/\\$/) {  # continuate line
# 	chop;
# 	chop;
# 	$l .= $_ ;
# 	redo lines ;
#     }

    if (/^:[ \t]*INCLUDE/) {
	s/^:[ \t]*INCLUDE[ \t]*//;
	chop ;
	unshift @filenames,$file;
	unshift @linesno,$lineno;

	$file=$f=$_;
        open ($file,$file) || die "No such file ".$file;
	$lineno = 0;
	&printlinenbr($lineno,$file) if($printlineno) ;
	redo lines;
    }

    }

    &printlinenbr($lineno) if($printlineno);

    if(eof($f)) {
	close($f) ;
	$file=$f=shift @filenames;
	$lineno=shift @linesno ;
	&printlinenbr($lineno,$file) if($printlineno);
    }

    $_ = $l . $_ ;
}


##MAIN
    $inperl=0;

    while(not eof(FILE)) {
	&getline(FILE);
      PROCESS: {
	if(/^:PERL{/) {
	    $_="\n";
	    $inperl=1;
	    last PROCESS;}
	if(/^:PERL}/) {
	    $_="\n";
	    $inperl=0;
	    last PROCESS;}
	if(/^:/) 
	{ 
	    s/:// ;
	    last PROCESS;
	}
	if(! $inperl) 	{    
	    &quote;
	    s/(.*)/print "$1\n";/;
	} 
    }

	print ;

    }

    # Then apply perl to this created file
    close(FILE);
    close(FILETEMP);
    open(FILEOUT,">$fileout")  ;
    select(FILEOUT);
    do $filetemp  ;
    unlink $filetemp unless ($keeptempfiles) ;


sub printlinenbr {
    # prints lines a la cpp
    local($nbr)=shift @_;
    local($name)=shift @_;

    print "#line $nbr  \"$name\"\n" ;
}

sub quote {
    # quotes a line so that it can print-ed unchanged by perl
    # woks by modifying $_

    # treat \
    s/\\/\\\\/g ;
    # but respect perl \l\u\L\U and \E escape sequences
    s/\\\\([ulUEL])/\\$1/g ;
    #take care of "
    s/\"/\\\"/g ;

    #for @array, we quote the @. The only problem is that
    #arrays can't be printed directly
    s/@/\\@/g;

    #$x[y] is incorectly interpreted -> we quote[
    s/\[/\\[/g;

    # and : (for :: constructs)
	    s/::/\\::/g;

    # For RCS/CVS protection, we need to embedd $foo$ keywords
    # this will be done with @$foo$@ mechanisms
    # (@ have already been qoted...)
    # @$ -> \$
    s/\\@\$/\\\$/g;
    # $@ -> \$
    s/\$\\@/\\\$/g;
}
