/*
 *  context.c
 *  
 *
 *  Created by nico on Sat Apr 07 2001.
 *  Copyright (c) 2001 Nicolas Sicard. All rights reserved.
 *
 *  $Id: context.c,v 1.1.1.1 2003/09/09 14:41:04 nico Exp $
 *  $Log: context.c,v $
 *  Revision 1.1.1.1  2003/09/09 14:41:04  nico
 *  Version 0.9
 *
 *  Revision 1.5  2003/07/20 21:06:50  nico
 *  Correction diverses apres compilation -Wall, virtualisation et reduction pour vnet (virtualnet.c)
 *
 *  Revision 1.4  2003/07/17 15:08:28  nico
 *  Reecriture multithread des fonctions WHERE et en partie ELSEWHERE
 *
 *  Revision 1.3  2003/07/07 21:52:25  nico
 *  Modification des keywords CVS
 *
 *  Revision 1.2  2003/07/07 20:23:32  nico
 *  Ajout de mot-cles CVS (auteur, revision, log)
 *
 */

#include "context.h"
#include "core.h"

ANET_CONTEXT_STACK *newContextStack(ANET_CORE *core)
{
  ANET_CONTEXT_STACK *stack = (ANET_CONTEXT_STACK*)malloc(sizeof(ANET_CONTEXT_STACK));

  c_init(stack, core);
  return stack;
}

/* allocate memory and initialize context object */
void c_init(ANET_CONTEXT_STACK *ctxs, ANET_CORE *core)
{
  if (ctxs == NULL) return;

  ctxs->depth = 0;
  ctxs->stack = (ANET_CONTEXT*)malloc(ANET_MAX_CTXDEPTH*sizeof(ANET_CONTEXT));
  ctxs->current = ctxs->stack;
  ctxs->size = AS_Size(core);
  ctxs->core = core;
  c_setEverywhere(ctxs->current, ctxs->size);
}

/* free context object allocated memory */
void c_free(ANET_CONTEXT_STACK *ctxs)
{
  if (ctxs == NULL) return;

  /* Warn if we detect unbalanced where/endwhere pairs */
  if (ctxs->depth != 0) {
	printf("context depth = %d\n", ctxs->depth);
    fprintf(stderr, "** ANET:CORE:CONTEXT> Warning: Unbalanced WHERE/ENDWHERE pairs.\n" );
  }
  /* Free memory space allocated for execution context */
  if (ctxs->stack != NULL) { free(ctxs->stack); ctxs->stack = NULL; }
  ctxs->current = NULL;
  ctxs->depth = 0;
}

/* set context object to EVERYWHERE */
void c_setEverywhere(ANET_CONTEXT *ctx, unsigned size)
{
  ctx->type = ctxtEverywhere;
  ctx->cond = NULL;
  ctx->size = size;
}

/* set context object to NOWHERE */
void c_setNowhere(ANET_CONTEXT *ctx)
{
  ctx->type = ctxtNowhere;
  ctx->cond = NULL;
  ctx->size = 0;
}

/* pop last context from the context stack */
void c_PopContext(ANET_CONTEXT_STACK *ctxs)
{
  ANET_CONTEXT *current = c_getCurrentContext(ctxs);
  if(c_isPvarContext(current)) {
    free(current->cond);
    current->cond = NULL;
    current->type = ctxtNowhere;
    current->size = 0;
  }
  ctxs->depth-- ;
  ctxs->current--;
}

typedef struct {
    unsigned char	*cond;
    //ANET_CORE		*core;
    unsigned char	*current, *previous;
    unsigned		size, flag;
} CONTEXT_PARAM;

/* c_Where task for multithreaded mode */
void *c_Where_Task(CONTEXT_PARAM *param)
{
    unsigned char	*current = param->current, *previous = param->previous;
    unsigned char	*cond = param->cond;

    if (previous != NULL){
        unsigned i;
        for (i = 0; i < param->size; i++)
            current[i] = (unsigned char)((int)previous[i] && (int)cond[i]);
    }else{ /* scan all "cond" to check active vertices */
        memcpy(current, cond, param->size*sizeof(unsigned char));
    }
    return NULL;
}

/* put a new context on top of the stack from boolean values */
void c_Where(ANET_CONTEXT_STACK *ctxs, unsigned char *cond)
{
  unsigned i, status;
  ANET_CONTEXT *current, *previous;

  // get adress of current context that will become the previous one
  previous = c_getCurrentContext(ctxs);
  
  c_PushContext(ctxs);
  current = c_getCurrentContext(ctxs);  // get new current context
  
  /* if previous context was "nowhere" then this is "nowhere" too ! */
  if (c_isNowhere(previous)){
    c_setNowhere(current);
    return;
  }
  
  current->cond = (unsigned char*)malloc(ctxs->size*sizeof(unsigned char));

  if (AS_Concurrency(ctxs->core) == 1) {	//1 thread

    // if previous context was from a boolean pvar then proceed to a
    // logical AND between "cond" and it
    if (c_isPvarContext(previous)){
        for (i = 0; i < ctxs->size; i++)
            current->cond[i] = (unsigned char)((int)previous->cond[i] && (int)cond[i]);
    }else{ /* scan all "cond" to check active vertices */
        memcpy(current->cond, cond, ctxs->size*sizeof(unsigned char));
    }

  } else {	//several threads

    pthread_t		task[ANET_MAX_CONCURRENCY];
    CONTEXT_PARAM	param[ANET_MAX_CONCURRENCY];
    
    for (i = 0; i < AS_BlockCount(ctxs->core); i++) {
        param[i].cond = (cond + AS_BlockStart(ctxs->core,i));
        param[i].current = (current->cond + AS_BlockStart(ctxs->core,i));
        if (c_isPvarContext(previous))
            param[i].previous = (previous->cond + AS_BlockStart(ctxs->core,i));
        else
            param[i].previous = NULL;
        param[i].size = AS_BlockSize(ctxs->core,i);

	status = pthread_create(&(task[i]), NULL, (void*(*)(void*))(c_Where_Task),(param+i));
	if (status != 0) printf("** ANET:CORE:CONTEXT(%s)> Error: Thread %d creation failed\n", "c_Where", i);
    }
    for (i = 0; i < AS_BlockCount(ctxs->core); i++) {
	status = pthread_join(task[i],NULL);
	if (status != 0) printf("** ANET:CORE:CONTEXT(%s)> Error: Thread %d ending failed\n", "c_Where", i);
    }
    
  }

  current->size = ctxs->size;
  current->type = ctxtPvarBool;
  if (current->size > 0) current->type = ctxtPvarBool; else c_setNowhere(current);
}

/* pop last context and make previous the current context */
void c_Endwhere(ANET_CONTEXT_STACK *ctxs) 
{
  c_PopContext(ctxs) ;
}

/* c_ElseWhere task for multithreaded mode */
void *c_ElseWhere_Task(CONTEXT_PARAM *param)
{
    unsigned char	*current = param->current, *previous = param->previous;
    unsigned		i;

    if(previous == NULL){ /* complement current context */
        for (i = 0; i < param->size; i++)
            current[i] = (unsigned char)(!(current[i]));
    }else{ /* combine previous and current contexts */
        for (i = 0; i < param->size; i++)
            current[i] = (unsigned char)(!(current[i]) && (int)(previous[i]));
    }
    return NULL;
}


/* replace current context by its complement -> ELSEWHERE */
void c_Elsewhere(ANET_CONTEXT_STACK *ctxs)
{
  unsigned i, status;
  ANET_CONTEXT *current, *previous = NULL;

  if (ctxs->depth == 0 || ctxs->depth > ANET_MAX_CTXDEPTH) {
    fprintf(stderr, "ANET> Error : No Where before Elsewhere or ANET_MAX_CTXDEPTH overflowed\n");
  }

  // get current and previous contexts
  current = c_getCurrentContext(ctxs);
  previous = c_getPreviousContext(ctxs);

  // if current context is EVERYWHERE
  if (c_isEverywhere(current)){
    fprintf(stderr,"ANET> Error : Elsewhere in everywhere context !\n");
    return;
  }

  // if current context is NOWHERE
  if (c_isNowhere(current)){
    if (c_isNowhere(previous)) return; // if previous context is nowhere, no change 
    if (c_isEverywhere(previous)) // if previous context is everywhere
      c_setEverywhere(current, ctxs->size);
    else{ // previous context is "somewhere"
      // then current context becomes previous context copy
      current->cond = (unsigned char*)malloc(sizeof(unsigned char)*ctxs->size);
      memcpy(current->cond, previous->cond, sizeof(unsigned char)*ctxs->size);
      current->type = previous->type;
      current->size = previous->size;
    }
    return;
  }

  if (AS_Concurrency(ctxs->core) == 1) {	//1 thread
  
    // current context is SOMEWHERE
    // if previous context is EVERYWHERE
    if(c_isEverywhere(previous)){ /* complement current context */
        for (i = 0; i < ctxs->size; i++)
            current->cond[i] = (unsigned char)(!(current->cond[i]));
    }else{ /* combine previous and current contexts */
        for (i = 0; i < ctxs->size; i++)
            current->cond[i] = (unsigned char)(!(current->cond[i]) && (int)(previous->cond[i]));
    }
    
  } else { //several threads
  
    pthread_t		task[ANET_MAX_CONCURRENCY];
    CONTEXT_PARAM	param[ANET_MAX_CONCURRENCY];
    
    for (i = 0; i < AS_BlockCount(ctxs->core); i++) {
        param[i].cond = NULL; //(cond + AS_BlockStart(ctxs->core,i));
        param[i].current = (current->cond + AS_BlockStart(ctxs->core,i));
        if (c_isEverywhere(previous))
            param[i].previous = NULL;
        else
            param[i].previous = (previous->cond + AS_BlockStart(ctxs->core,i));
        param[i].size = AS_BlockSize(ctxs->core,i);

	status = pthread_create(&(task[i]), NULL, (void*(*)(void*))(c_ElseWhere_Task),(param+i));
	if (status != 0) printf("** ANET:CORE:CONTEXT(%s)> Error: Thread %d creation failed\n", "c_Where", i);
    }
    for (i = 0; i < AS_BlockCount(ctxs->core); i++) {
	status = pthread_join(task[i],NULL);
	if (status != 0) printf("** ANET:CORE:CONTEXT(%s)> Error: Thread %d ending failed\n", "c_Where", i);
    }
    
  }
}
