/*
 *  context.h
 *  
 *
 *  Created by nico on Sat Apr 07 2001.
 *  Copyright (c) 2001 IEF. All rights reserved.
 *
 *  $Id: context.h,v 1.1.1.1 2003/09/09 14:41:04 nico Exp $
 *  $Log: context.h,v $
 *  Revision 1.1.1.1  2003/09/09 14:41:04  nico
 *  Version 0.9
 *
 *  Revision 1.4  2003/07/17 15:08:28  nico
 *  Reecriture multithread des fonctions WHERE et en partie ELSEWHERE
 *
 *  Revision 1.3  2003/07/07 21:52:25  nico
 *  Modification des keywords CVS
 *
 *  Revision 1.2  2003/07/07 20:23:32  nico
 *  Ajout de mot-cles CVS (auteur, revision, log)
 *
 */

#ifndef _CONTEXT_H_
#define _CONTEXT_H_

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* maximum number of nested context */
#define ANET_MAX_CTXDEPTH 		32

#define ANET_WOCONTEXT			0x00001000

struct ANET_CORE;

typedef struct {
  unsigned char *cond;
  unsigned size;
  int type;
} ANET_CONTEXT;

/* low level structure of context */
typedef struct {
  unsigned		depth, size;
  ANET_CONTEXT		*stack;
  ANET_CONTEXT		*current;
  struct ANET_CORE	*core;
} ANET_CONTEXT_STACK;

ANET_CONTEXT_STACK *newContextStack(struct ANET_CORE *core);
#define freeContextStack 		c_free

#define CONTEXT_WHERE			c_Where
#define CONTEXT_ENDWHERE		c_Endwhere
#define CONTEXT_ELSEWHERE		c_Elsewhere

#define CONTEXT_INIT 			c_init
#define CONTEXT_FREE			c_free
#define CONTEXT_SET_EVERYWHERE		c_setEverywhere
#define CONTEXT_SET_NOWHERE 		c_setNowhere
#define CONTEXT_PUSH			c_PushContext
#define CONTEXT_POP			c_PopContext
#define CONTEXT_IS_EVERYWHERE		c_isEverywhere
#define CONTEXT_WAS_EVERYWHERE		c_wasEverywhere
#define CONTEXT_IS_NOWHERE		c_isNowhere
#define CONTEXT_WAS_NOWHERE 		c_wasNowhere
#define CONTEXT_IS_PVAR			c_isPvarContext
#define CONTEXT_WAS_PVAR		c_wasPvarContext
#define CONTEXT_GET_NODES		c_getContextNodes

enum { ctxtEverywhere, ctxtNowhere, ctxtPvarBool };

/* context stack initialization and destruction */
void c_init(ANET_CONTEXT_STACK *ctxs, struct ANET_CORE *core);
void c_free(ANET_CONTEXT_STACK *ctxs);

/* internal context initialisations */
void c_setEverywhere(ANET_CONTEXT *ctx, unsigned size);
void c_setNowhere(ANET_CONTEXT *ctx);

/* internal low-level context manipulations and acessors */
#define c_PushContext(ctxs) { (ctxs)->depth++ ; (ctxs)->current++; }
/*
void c_PushContext(ANET_CONTEXT_STACK *ctxs)
{
  ctxs->depth++;
  (ctxs)->current++;
}
*/
void c_PopContext(ANET_CONTEXT_STACK *);

#define c_isEverywhere(ctx) ((ctx)==NULL || (ctx)->type==ctxtEverywhere)
#define c_wasEverywhere(ctxs,i) ((ctxs)->stack[(ctxs)->depth+(i)].type == ctxtEverywhere)
/*
int c_isEverywhere(ANET_CONTEXT *ctx)
{
  return ((ctx)==NULL || (ctx)->type==ctxtEverywhere);
}
*/
#define c_isPvarContext(ctx) ((ctx) != NULL && (ctx)->type==ctxtPvarBool)
#define c_wasPvarContext(ctxs,i) ((ctxs)->stack[(ctxs)->depth+(i)].type == ctxtPvarBool)
/*
int c_isPvarContext(ANET_CONTEXT *ctx)
{
  return ((ctx) != NULL && (ctx)->type==ctxtPvarBool);
}
*/
#define c_isNowhere(ctx) ((ctx) != NULL && (ctx)->type==ctxtNowhere)
#define c_wasNowhere(ctxs,i) ((ctxs)->stack[(ctxs)->depth+(i)].type == ctxtNowhere)

/*
int c_isNowhere(ANET_CONTEXT *ctx){
  return ((ctx) != NULL && (ctx)->type==ctxtNowhere);
}
*/

#define c_getContext(ctxs,i) (((ctxs)->depth < ANET_MAX_CTXDEPTH)? ((ctxs)->stack[(ctxs)->depth+(i)]): NULL)
#define c_getCurrentContext(ctxs) ((ctxs)->current)
#define c_getPreviousContext(ctxs) ((ctxs)->current - 1)

/* Low level methods (or macros) to get context stack memory space and size */
#define c_getContextNodes(ctx) (((ctx) != NULL)? (ctx)->cond : NULL)
/*
unsigned *c_getContextNodes(ANET_CONTEXT *ctx) {
  return (((ctx) != NULL)? (ctx)->indices : NULL);
}
*/
#define c_getContextSize(ctx) (((ctx) != NULL)? (ctx)->size : 0)
/*
unsigned c_getContextSize(ANET_CONTEXT *ctx)
{
  return (((ctx) != NULL)? (ctx)->size : 0);
}
*/

/* context user fonctions (for C++ methods) */
#define c_Everywhere(ctxs) \
  { c_PushContext(ctxs); c_setEverywhere(c_getCurrentContext(ctxs), AS_Size((ctxs)->core)); }
/*
void c_Everywhere(ANET_CONTEXT_STACK *ctxs) 
{
  c_PushContext(ctxs);
  c_setEverywhere(c_getCurrentContext(ctxs));
}
*/
void c_Where(ANET_CONTEXT_STACK *ctx, unsigned char *cond);
void c_Endwhere(ANET_CONTEXT_STACK *ctx);
void c_Elsewhere(ANET_CONTEXT_STACK *ctx);

#endif /* _CONTEXT_H_ */
