/*
 *  core.c
 *  Anet
 *
 *  Created by Nicolas Sicard on Tue Sep 03 2002.
 *  Copyright (c) 2002 IEF. All rights reserved.
 *
 *  $Id: core.c,v 1.1.1.1 2003/09/09 14:41:04 nico Exp $
 *  $Log: core.c,v $
 *  Revision 1.1.1.1  2003/09/09 14:41:04  nico
 *  Version 0.9
 *
 *  Revision 1.7  2003/07/17 15:08:28  nico
 *  Reecriture multithread des fonctions WHERE et en partie ELSEWHERE
 *
 *  Revision 1.6  2003/07/17 12:11:11  nico
 *  *** empty log message ***
 *
 *  Revision 1.5  2003/07/15 13:24:22  nico
 *  Correction du bogue d'incoherence de la topologie des blocs apres changement de nombre de threads en cours de programme (modif des blocs placee dans setConcurrency)
 *
 *  Revision 1.4  2003/07/07 21:52:25  nico
 *  Modification des keywords CVS
 *
 *  Revision 1.3  2003/07/07 21:30:22  nico
 *  Ajout des fonctions getCoreConcurrency et setCoreConcurrency
 *
 *  Revision 1.2  2003/07/07 20:23:32  nico
 *  Ajout de mot-cles CVS (auteur, revision, log)
 *
 */

#include "core.h"

static ANET_CORE *gCore = NULL;

ANET_CORE *sharedCore(ANET_TOPOLOGY *topology, unsigned int concurrency)
{
  if (topology == NULL && gCore == NULL) {
	fprintf(stderr, "** ANET:CORE:CORE> Fatal Error: Invalid topology. Please check topology.h.\n");
	return NULL;
  }
  if (gCore != NULL) return gCore;
  // if global object is not already allocated, do it
  gCore = (ANET_CORE*)malloc(sizeof(ANET_CORE));

  // fill data structure
  gCore->topology = topology;
  gCore->context_stack = newContextStack(gCore);
  setCoreConcurrency(gCore, concurrency);
  
#ifdef ANET_NET
  gCore->client = newClient();
  if (connectToServer(gCore->client) == -1) {
	fprintf(stderr, "** ANET:CORE:CORE> Error: No server found!\n");
	gCore->client = freeClient(gCore->client);
  } else {
	if (registerClient(gCore->client, AS_Size(gCore)*8)) {
	  fprintf(stderr, "** ANET:CORE:CORE> Error: Registration failed!\n");
	}
  }
#endif /* ANET_NET */

  ANET_PRINT_STAMP;
  return gCore;
}

void closeCore(ANET_CORE *core)
{
  if (gCore == NULL) return;
  freeContextStack(gCore->context_stack);
#ifdef ANET_NET
  gCore->client = freeClient(gCore->client);
#endif /* ANET_NET */  
  free(gCore); gCore = NULL;
}

// return current concurrency
unsigned getCoreConcurrency(ANET_CORE *core)
{
    return AS_Concurrency(core);
}

// set current concurrency
void setCoreConcurrency(ANET_CORE *core, unsigned int concurrency)
{
    unsigned int temp = concurrency;
    unsigned int lines, i;

    // fix dummy concurrency
    if (concurrency < 1) concurrency = 1;
    if (concurrency > ANET_MAX_CONCURRENCY) concurrency = ANET_MAX_CONCURRENCY;
    if (concurrency > AS_Topology(core)->height) concurrency = AS_Topology(core)->height;
    if (concurrency != temp) // parameter has been changed
	fprintf(stderr, "** ANET:CORE:CORE> Warning: Concurrency changed to %d.\n", concurrency);
    core->concurrency = concurrency;
    
    // build block distribution info from concurrency and image dimensions
    lines = AS_Topology(core)->height / concurrency;
    for (i = 0; i < concurrency; i++) {
        core->block[i] = i * (lines * AS_Topology(core)->width);
    }
    core->block[i] = core->topology->size;

}


// allocate and initialize a new sync barrier
ANET_BARRIER *new_barrier()
{
    ANET_BARRIER *barrier = (ANET_BARRIER*)malloc(sizeof(ANET_BARRIER));
    barrier->instance = 0;
    barrier->blocked_threads = 0;
    pthread_mutex_init(&(barrier->sync_lock), NULL);
    pthread_cond_init(&(barrier->sync_cond), NULL);

    return barrier;
}
