/*
 *  core.h
 *  Anet
 *
 *  Created by Nicolas Sicard on Tue Sep 03 2002.
 *  Copyright (c) 2002 IEF. All rights reserved.
 *
 *  $Id: core.h,v 1.4 2003/12/08 18:43:35 nico Exp $
 *  $Log: core.h,v $
 *  Revision 1.4  2003/12/08 18:43:35  nico
 *  Ajout du masque ANET_ASYNC pour borgefor parallele asynchrone
 *
 *  Revision 1.3  2003/10/24 15:43:01  nico
 *  Ajout duflag ANET_DIFUSION
 *
 *  Revision 1.2  2003/09/10 09:20:16  nico
 *  Renommage de mtrgraph en rgraph
 *
 *  Revision 1.1.1.1  2003/09/09 14:41:04  nico
 *  Version 0.9
 *
 *  Revision 1.6  2003/07/24 14:32:38  nico
 *  Retour a la version optimisee
 *
 *  Revision 1.5  2003/07/20 21:06:50  nico
 *  Correction diverses apres compilation -Wall, virtualisation et reduction pour vnet (virtualnet.c)
 *
 *  Revision 1.4  2003/07/07 21:52:25  nico
 *  Modification des keywords CVS
 *
 *  Revision 1.3  2003/07/07 21:30:22  nico
 *  Ajout des fonctions getCoreConcurrency et setCoreConcurrency
 *
 *  Revision 1.2  2003/07/07 20:23:32  nico
 *  Ajout de mot-cles CVS (auteur, revision, log)
 *
 */

#ifndef _CORE_H_
#define _CORE_H_

#include "topology.h"
#include "context.h"
#ifdef ANET_NET
#include "client.h"
#endif /* ANET_NET */

#ifndef ANET_VERS
#define ANET_VERS "Unknown"
#endif

#ifndef ANET_SYS
#define ANET_SYS  "Unknown"
#endif

#ifndef ANET_PROC
#define ANET_PROC "Unknown"
#endif

#define ANET_PRINT_STAMP   printf("Anet Library v%s - %s,%s\n", ANET_VERS, __DATE__, __TIME__);


// Operation modifiers
#define ANET_EXTENDED			0x00000100
#define ANET_EXCLUSIVE			0x00000200
#define ANET_INVERT				0x00000400
#define ANET_SYMETRIZE			0x00000800
#define ANET_WOCONTEXT			0x00001000
#define ANET_WITHCONTEXT		0x00002000
#define ANET_TWOPASS			0x00004000
#define ANET_WITHRNET			0x00008000
#define ANET_UNKNOWN			0x00010000
#define ANET_DIFFUSION			0x00020000
#define ANET_SORTRNET			0x00040000
#define ANET_ASYNC				0x00080000


#define ANET_MAX_CONCURRENCY	64

/* Anet core object */
typedef struct ANET_CORE {
  unsigned int 			concurrency;			// number of tasks
  ANET_TOPOLOGY 		*topology;			// pvar topology
  ANET_CONTEXT_STACK		*context_stack;			// contexts
  unsigned int			block[ANET_MAX_CONCURRENCY+1];	// pvar block distribution
#ifdef ANET_NET
  ANET_CLIENT			*client;
#endif /* ANET_NET */
} ANET_CORE;


ANET_CORE *sharedCore(ANET_TOPOLOGY *topology, unsigned int concurrency);
void closeCore(ANET_CORE *core);
void setCoreConcurrency(ANET_CORE *core, unsigned int concurrency);
unsigned getCoreConcurrency(ANET_CORE *core);

/*
 Synchronisation barrier data structure
 */
typedef struct ANET_BARRIER {
    unsigned int 	instance, blocked_threads;
    pthread_mutex_t	sync_lock;
    pthread_cond_t	sync_cond;
} ANET_BARRIER;

ANET_BARRIER *new_barrier();

// Some very useful macros !!
#define AS_Concurrency(core) ((core)->concurrency)
#define AS_Topology(core) ((core)->topology)
#define AS_Width(core) (AT_Width((core)->topology))
#define AS_Height(core) (AT_Height((core)->topology))
#define AS_Size(core) (AT_Size((core)->topology))
#define AS_ContextStack(core) ((core)->context_stack)
#define AS_Context(core) (c_getCurrentContext((core)->context_stack))
#define AS_ContextNodes(core) (CONTEXT_GET_NODES(AS_Context(core)))

#define AS_Error(message) fprintf(stderr, "ANET> Error : %s\n", "message");

// Get number of blocks
#define AS_BlockCount(core) ((core)->concurrency)
// Get last block index
#define AS_LastBlock(core) ((core)->concurrency-1)
// Get first pixel index of block b
#define AS_BlockStart(core,b) ((core)->block[(b)])
// Get number of pixels for block b
#define AS_BlockSize(core, b) ((core)->block[(b+1)] - (core)->block[(b)])
// Print first two blocks
#define AS_Print(core) { printf("%d %d %d %d\n", (core)->width, (core)->height, (core)->size, (core)->concurrency); printf("%d %d\n",(core)->block[0],(core)->block[1]);}

// Parameter acessors
#define AP_Res(param) ((param)->res)
#define AP_Src(i, param) ((param)->src ## i)
#define AP_Buffer(param) ((param)->buffer)
#define AP_Mask(param) ((param)->mask)
#define AP_Context(param) ((param)->ctxt)
#define AP_Block(param) ((param)->block)
#define AP_Size(param) ((param)->size)
#define AP_Core(param) ((param)->core)
#define AP_Flag(param) ((param)->flag)

#define AP_BlockStart(param) (AS_BlockStart((param)->core,(param)->block))
#define AP_BlockSize(param) (AS_BlockSize((param)->core,(param)->block))
#define AP_BlockCount(param) (AS_BlockCount((param)->core))
#define AP_Width(param) AS_Width(AP_Core(param))
#define AP_Height(param) AS_Height(AP_Core(param))
#define AP_TaskCount(param) AS_Concurrency(AP_Core(param))

#define CORE(r) (r)->core
#define SCC(r) (r)->scc

#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>

#define NEW_TIME(t) \
struct timezone t ## _tz; \
struct timeval t ## _tv, t ## _tstart;

#define START_TIME(t) \
gettimeofday(& t ## _tstart, & t ## _tz);

#define END_TIME(t) \
gettimeofday(& t ## _tv, & t ## _tz);

#define PRINT_TIME(t,msg) \
printf("%s = %f ms\n", msg, (float)( t ## _tv.tv_usec - t ## _tstart.tv_usec \
                                     + 1000000*( t ## _tv.tv_sec - t ## _tstart.tv_sec)) / 1000.0)

#endif /* _CORE_H_ */
