/*
 *  mgraph.c
 *  Anet
 *
 *  Created by Nicolas Sicard on Thu Sep 19 2002.
 *  Copyright (c) 2002 IEF. All rights reserved.
 *
 *  $Id: mgraph.c,v 1.1.1.1 2003/09/09 14:41:04 nico Exp $
 *  $Log: mgraph.c,v $
 *  Revision 1.1.1.1  2003/09/09 14:41:04  nico
 *  Version 0.9
 *
 *  Revision 1.1  2003/07/25 19:33:06  nico
 *  renommage de invert.c et invert.h en mgraph.c et mgraph.h
 *
 *  Revision 1.5  2003/07/25 12:16:19  nico
 *  Remplacement de bzero par memset
 *
 *  Revision 1.4  2003/07/20 21:06:50  nico
 *  Correction diverses apres compilation -Wall, virtualisation et reduction pour vnet (virtualnet.c)
 *
 *  Revision 1.3  2003/07/07 21:52:25  nico
 *  Modification des keywords CVS
 *
 */

#include "mgraph.h"

void INVERT_SUBNET_TASK(INVERT_PARAM *p);

void INVERT_SUBNET(ANET_MASK *res, ANET_MASK *src, ANET_CORE *core, int flag)
{
    pthread_t 		task[ANET_MAX_CONCURRENCY];
    INVERT_PARAM 	param[ANET_MAX_CONCURRENCY];
    unsigned 		i, status;
	ANET_MASK		*lres = res;

    // allocate buffer if source and destination are the same !
    if (res == src) lres = (ANET_MASK *)malloc(sizeof(ANET_MASK)*AS_Size(core));
	memset(lres, 0x00, sizeof(ANET_MASK)*AS_Size(core));

    for (i = 0; i < AS_BlockCount(core); i++) {
        AP_Res(param+i) = lres;
        AP_Src(1,param+i) = src;
        AP_Block(param+i) = i;
        AP_Size(param+i) = AS_Size(core); //AS_BlockSize(core,i);
        AP_Core(param+i) = core;
        AP_Flag(param+i) = flag;

        status = pthread_create(&(task[i]), NULL, (void*(*)(void*))(INVERT_SUBNET_TASK),&(param[i]));
        if (status != 0) printf("** ANET:CORE:INVERT> Warning: Thread %d creation failed\n", i);
    }
    for (i = 0; i < AS_BlockCount(core); i++) {
        status = pthread_join(task[i],NULL);
        if (status != 0) printf("** ANET:CORE:INVERT> Warning: Thread %d ending failed\n", i);
    }
	if (lres != res) {
	  memcpy(res, lres, sizeof(ANET_MASK)*AS_Size(core));
	  free(lres);
	}
}

void INVERT_SUBNET_TASK(INVERT_PARAM *param)
{
  ANET_MASK *res = AP_Res(param);
  ANET_MASK *src = AP_Src(1,param);
  int sym = (AP_Flag(param) & ANET_SYMMETRIZE);
  ANET_MASK mtmp;
  unsigned lastnode, i, cur_node = 0;
  unsigned width = AP_Width(param), size = AP_Size(param);

  if (AP_Block(param) == 0) {	// process first block
    /** first line **/
    /* node 0 (first node of first line) */
    mtmp = src[0];
    if (mtmp != 0x00){
	  if (mtmp & 1)  { res[size-width] |= 16; }
	  if (mtmp & 2)  { res[size-width+1] |= 32; }
	  if (mtmp & 4)  { res[1] |= 64; }
	  if (mtmp & 8)  { res[width+1] |= 128; }
	  if (mtmp & 16) { res[width] |= 1; }
	  if (mtmp & 32) { res[width+width-1] |= 2; }
	  if (mtmp & 64) { res[width-1] |= 4; }
	  if (mtmp & 128){ res[size-1] |= 8; }
	}
	cur_node++;
    /* nodes 1 -> width-2 */
    for (i = 1; i < width-1; i++){
	  mtmp = src[cur_node];
	  if (mtmp & 1)  { res[cur_node+size-width] |= 16; }
	  if (mtmp & 2)  { res[cur_node+size-width+1] |= 32; }
	  if (mtmp & 4)  { res[cur_node+1] |= 64; }
	  if (mtmp & 8)  { res[cur_node+width+1] |= 128; }
	  if (mtmp & 16) { res[cur_node+width] |= 1; }
	  if (mtmp & 32) { res[cur_node+width-1] |= 2; }
	  if (mtmp & 64) { res[cur_node-1] |= 4; }
	  if (mtmp & 128){ res[cur_node+size-width-1] |= 8;}
	  cur_node++;
    }
    /* node width-1 (last node of first line) */
	mtmp = src[cur_node];
	if (mtmp != 0x00){
	  if (mtmp & 1)  { res[size-1] |= 16; }
	  if (mtmp & 2)  { res[size-width] |= 32; }
	  if (mtmp & 4)  { res[0] |= 64; }
	  if (mtmp & 8)  { res[width] |= 128; }
	  if (mtmp & 16) { res[width+width-1] |= 1; }
	  if (mtmp & 32) { res[width+width-2] |= 2; }
	  if (mtmp & 64) { res[width-2] |= 4; }
	  if (mtmp & 128){ res[size-2] |= 8; }
	}
	cur_node++;

	// startline = 1;
  } else {
    cur_node = AP_BlockStart(param);
  }

  /* lines 1 -> height-2 */
  if (AP_Block(param) < AS_LastBlock(AP_Core(param))) {
	lastnode = AP_BlockStart(param) + AP_BlockSize(param);
  } else
	lastnode = size - width;
  //printf("curnode=%d, lastnode=%d\n",cur_node, lastnode);
  while (cur_node < lastnode){
	/* first node */
	mtmp = src[cur_node];
	if (mtmp != 0x00){
	  if (mtmp & 1)  { res[cur_node-width] |= 16; }
	  if (mtmp & 2)  { res[cur_node-width+1] |= 32; }
	  if (mtmp & 4)  { res[cur_node+1] |= 64; }
	  if (mtmp & 8)  { res[cur_node+width+1] |= 128; }
	  if (mtmp & 16) { res[cur_node+width] |= 1; }
	  if (mtmp & 32) { res[cur_node+width+width-1] |= 2; }
	  if (mtmp & 64) { res[cur_node+width-1] |= 4; }
	  if (mtmp & 128){ res[cur_node-1] |= 8; }
	}
	//printf("res[%d]=%d\n", cur_node+width, res[cur_node+width]);
	cur_node++;

	/* nodes 1 -> width-2 */
	for (i = 1; i < width-1; i++){
	  mtmp = src[cur_node];
	  if (mtmp & 1)  { res[cur_node-width] |= 16; }
	  if (mtmp & 2)  { res[cur_node-width+1] |= 32; }
	  if (mtmp & 4)  { res[cur_node+1] |= 64; }
	  if (mtmp & 8)  { res[cur_node+width+1] |= 128; }
	  if (mtmp & 16) { res[cur_node+width] |= 1; }
	  if (mtmp & 32) { res[cur_node+width-1] |= 2; }
	  if (mtmp & 64) { res[cur_node-1] |= 4; }
	  if (mtmp & 128){ res[cur_node-width-1] |= 8; }
	  cur_node++;
	}
	
	/* last node */
	mtmp = src[cur_node];
	if (mtmp!= 0x00){
	  if (mtmp & 1)  { res[cur_node-width] |= 16; }
	  if (mtmp & 2)  { res[cur_node-width-width+1] |= 32; }
	  if (mtmp & 4)  { res[cur_node-width+1] |= 64; }
	  if (mtmp & 8)  { res[cur_node+1] |= 128; }
	  if (mtmp & 16) { res[cur_node+width] |= 1; }
	  if (mtmp & 32) { res[cur_node+width-1] |= 2; }
	  if (mtmp & 64) { res[cur_node-1] |= 4; }
	  if (mtmp & 128){ res[cur_node-width-1] |= 8; }
	}
	cur_node++;
  }

  /* last line (only if processed block is last one) */
  if (AP_Block(param) == AS_LastBlock(AP_Core(param))) {
	cur_node = size - width;
	mtmp = src[cur_node];
	if (mtmp != 0x00){
	  if (mtmp & 1)  { res[cur_node-width] |= 16; }
	  if (mtmp & 2)  { res[cur_node-width+1] |= 32; }
	  if (mtmp & 4)  { res[cur_node+1] |= 64; }
	  if (mtmp & 8)  { res[1] |= 128; }
	  if (mtmp & 16) { res[0] |= 1; }
	  if (mtmp & 32) { res[width-1] |= 2; }
	  if (mtmp & 64) { res[size-1] |= 4; }
	  if (mtmp & 128){ res[cur_node-1] |= 8; }
	}
	cur_node++;
	
	/* nodes 1 -> width-2 */
	for (i = 1; i < width-1; i++){
	  mtmp = src[cur_node];
	  if (mtmp & 1)  { res[cur_node-width] |= 16; }
	  if (mtmp & 2)  { res[cur_node-width+1] |= 32; }
	  if (mtmp & 4)  { res[cur_node+1] |= 64; }
	  if (mtmp & 8)  { res[cur_node+width-size+1] |= 128; }
	  if (mtmp & 16) { res[cur_node+width-size] |= 1; }
	  if (mtmp & 32) { res[cur_node+width-size-1] |= 2; }
	  if (mtmp & 64) { res[cur_node-1] |= 4; }
	  if (mtmp & 128){ res[cur_node-width-1] |= 8; }
	  cur_node++;
	}
	
	/* last node */
	mtmp = src[cur_node];
	if (mtmp!= 0x00){
	  if (mtmp & 1)  { res[cur_node-width] |= 16; }
	  if (mtmp & 2)  { res[cur_node-width-width+1] |= 32; }
	  if (mtmp & 4)  { res[cur_node-width+1] |= 64; }
	  if (mtmp & 8)  { res[0] |= 128; }
	  if (mtmp & 16) { res[width-1] |= 1; }
	  if (mtmp & 32) { res[width-2] |= 2; }
	  if (mtmp & 64) { res[cur_node-1] |= 4; }
	  if (mtmp & 128){ res[cur_node-width-1] |= 8; }
	}
	cur_node++;
  }
  if (sym & ANET_SYMMETRIZE){
	  for (i = 0; i < AP_Size(param); i++) res[i] |= src[i];
  }

}

void MASK_SUBNET(ANET_MASK *src, ANET_CORE *core, int mask_id)
{
    unsigned width = AS_Width(core), size = AS_Size(core), height = AS_Height(core);
    unsigned c,l;

	if (mask_id == Torus) {
		for(c = 0; c < width; c++) {
			src[c] &= ~(ANET_MASK)(mN|mNW|mNE) ;
			src[size-c-1] &= ~(ANET_MASK)(mS|mSW|mSE) ;
		}
		for(l = 0; l < height; l++) {
			src[width*l] &= ~(ANET_MASK)(mW|mNW|mSW) ;
			src[width*(l+1)-1] &= ~(ANET_MASK)(mE|mNE|mSE) ;
		}
	}
}
