/*
 *  pgm.c
 *  
 *
 *  Created by nico on Fri May 04 2001.
 *  Copyright (c) 2001 IEF. All rights reserved.
 *
 *  $Id: pgm.c,v 1.3 2003/12/08 18:42:55 nico Exp $
 *  $Log: pgm.c,v $
 *  Revision 1.3  2003/12/08 18:42:55  nico
 *  Correction bogue ecriture pgm binaire P5 (format de fichier pas exactement respecte)
 *
 *  Revision 1.2  2003/09/11 23:10:46  nico
 *  Ajout de la lecture/ecriture de fichiers PGM binaires
 *
 *  Revision 1.1.1.1  2003/09/09 14:41:04  nico
 *  Version 0.9
 *
 *  Revision 1.6  2003/09/01 16:17:03  nico
 *  Correction de transtypages implicites unsigned-signed int
 *
 *  Revision 1.5  2003/07/20 21:06:50  nico
 *  Correction diverses apres compilation -Wall, virtualisation et reduction pour vnet (virtualnet.c)
 *
 *  Revision 1.4  2003/07/08 11:34:49  nico
 *  Ajout de la fonction readPgmSizeFromPath
 *
 *  Revision 1.3  2003/07/07 21:52:25  nico
 *  Modification des keywords CVS
 *
 *  Revision 1.2  2003/07/07 20:23:32  nico
 *  Ajout de mot-cles CVS (auteur, revision, log)
 *
 */

#include "pgm.h"
#include <stdlib.h>
#include <ctype.h>

int eat_spaces(FILE *file)
{
	int c;
	do { c = getc(file); } while (isspace(c));
	return c;
}

int get_next_value(FILE *file)
{
    unsigned int c, value = 0;
    c = eat_spaces(file);
    //eat comments if exist
    while (c == '#'){
        do { c = getc(file); } while (c != '\n' && c != '\r');
        c = eat_spaces(file);
    }
    if (c < '0' || c > '9') return -1;
    while ( (c >= '0') && (c <= '9') && (value < 0xffffffff / 10 - 1) ) {
        value = 10*value + (c - (int)'0');
        c = getc(file);
    }
    return value;
}   

int readPgmHeaderFromStream(FILE *file, unsigned int *width, unsigned int *height, unsigned short *type)
{
    int c;
    if (file == NULL || width == NULL || height == NULL) return PGM_PARAM_ERROR;

    //check magick number
    c = eat_spaces(file);
    if (c != 'P') return PGM_FILE_TYPE_ERROR;
    c = eat_spaces(file);
    if (type) *type = (unsigned short)(c -'0');
    if (c != '2' && c != '3' && c != '5') return PGM_FILE_TYPE_ERROR;
    //check width, height and maximum gray value
    *width = get_next_value(file);
    if (*width < 1) return PGM_FILE_FORMAT_ERROR;
    *height = get_next_value(file);
    if (*height < 1) return PGM_FILE_FORMAT_ERROR;
    if (get_next_value(file) == -1) return PGM_FILE_FORMAT_ERROR;
    return 0;
}

int readPgmFromStream(FILE *file, unsigned char *buffer, unsigned int size)
{
    int c, i, err;
    unsigned int width, height, count;
    unsigned short  type;
    
    //check parameters
    if (file == NULL || buffer == NULL) return PGM_PARAM_ERROR;
    if ((err = readPgmHeaderFromStream(file, &width, &height, &type))) return err;

    count = (size < width*height)? size : width*height;
    //get values
    if (type == PGM_BINARY) {
        if (fread(buffer, sizeof(unsigned char), count, file) < count)
            return PGM_MEMORY_ERROR;
    } else {
        for (i = 0; i < count; i++){
            c = get_next_value(file);
            if (c < 0) return PGM_FILE_FORMAT_ERROR;
            *buffer = (unsigned char)c; buffer++;
        }
    }
    return PGM_NOERR;
}

int writePgmToStream(FILE *file, const unsigned char *buffer, unsigned int width,
                     unsigned int height, unsigned short type, int mod)
{
    int i;
    //check parameters
    if (file == NULL || buffer == NULL) return PGM_PARAM_ERROR;

    //write values
    fprintf(file, "P%d\n", type);
    fprintf(file, "#created by Anet library\n");
    fprintf(file,  "%d %d\n255\n", width, height);
    if (type == PGM_BINARY) {
        if (fwrite(buffer, sizeof(unsigned char), width*height, file) < width*height)
            return PGM_MEMORY_ERROR;        
    } else {
        for (i = 0; i < width*height; i++){
            if (i % mod == 0) fprintf(file, "\n%3d", (int)buffer[i]);
            else fprintf(file, " %3d", (int)buffer[i]);
            //printf("%d ", (int)buffer[i]);
        }
    }
    return PGM_NOERR;
}
/*
int readAndAllocPgmFromStream(FILE *file, unsigned char **buffer, int *width, int *height)
{
	int c, i, err;
	int gmax;
	
	//check parameters
	if (file == NULL || buffer == NULL || width == NULL || height == NULL) return PGM_PARAM_ERROR;
	if (err = readPgmHeaderFromStream(file, width, height)) return err;
	
	*buffer = (unsigned char)malloc((*width)*(*height)*sizeof(unsigned char));
	for (i = 0; i < (*height)*(*width); i++){
		c = get_next_value(file);
		if (c < 0) { free(*buffer); return PGM_FILE_FORMAT_ERROR; }
		(*buffer)[i] = (unsigned char)c;
	}
	return 0;
}
*/

int readPgmSizeFromPath(const char *path, unsigned int *width, unsigned int *height)
{
    FILE *file;
    int err;
    
    file = fopen(path,"r");
    if (file == NULL) return PGM_OPENFILE_ERROR;
    err = readPgmHeaderFromStream(file, width, height, NULL);
    fclose(file);
    return err;
}

int readPgmFromPath(const char *path, unsigned char *buffer, int size)
{
    FILE *file;
    int err;
    
    file = fopen(path,"r");
    if (file == NULL) return PGM_OPENFILE_ERROR;
    err = readPgmFromStream(file,buffer,size);
    fclose(file);
    return err;
}

int writePgmToPath(const char *path, const unsigned char *buffer, int width, int height, unsigned short type)
{
    FILE *file;
    int err;

    if (path != NULL) {
        file = fopen(path,"w");
        if (file == NULL) return PGM_CREATEFILE_ERROR;
    } else {
        file = stdout;
    }
    err = writePgmToStream(file,buffer,width,height, type, 16);
    if (file != stdout)
        fclose(file);
    else
        printf("\n");
    return err;
}
