/*
 *  spanning.c
 *  Anet
 *
 *  Created by nico on Wed Nov 07 2001.
 *  Copyright (c) 2001 IEF. All rights reserved.
 *
 *  $Id: spanning.c,v 1.1.1.1 2003/09/09 14:41:04 nico Exp $
 *  $Log: spanning.c,v $
 *  Revision 1.1.1.1  2003/09/09 14:41:04  nico
 *  Version 0.9
 *
 *  Revision 1.3  2003/07/07 21:52:25  nico
 *  Modification des keywords CVS
 *
 *  Revision 1.2  2003/07/07 20:23:32  nico
 *  Ajout de mot-cles CVS (auteur, revision, log)
 *
 */

#include "spanning.h"

#ifndef RAND_MAX
#define RAND_MAX (0x7fffffff - 1)
#endif
#define random() rand()
#define srandom(i) srand(i)

enum {WHITE, BLACK, GREY};

typedef struct _ANET_FIFO {
	unsigned *tab;
	int size, start, end, nbr;
} fifo;

#define fifo_alloc(f,n) { (f).tab = (unsigned*)malloc((n)*sizeof(unsigned));\
(f).start = (f).end = (f).nbr = 0; (f).size = n; }
#define fifo_free(f) { if ((f).tab) free((f).tab); }
#define fifo_init(f) { (f).start = (f).end = (f).nbr = 0; }
#define fifo_insert(f,i) { (f).tab[(f).end] = i; (f).nbr++; if (++(f).end == (f).size) (f).end = 0; }
#define fifo_notempty(f) ((f).nbr)

unsigned fifo_extract(fifo *f) {
    unsigned x = f->tab[f->start];
    f->nbr--;
    if (++(f->start) == f->size) f->start = 0;
    return x;
}

void SPANNING_TREE(ANET_MASK *res, ANET_MASK *src, Ubyte *seeds, float rnd, ANET_CORE *core)
{
    register unsigned next_node;
    unsigned neighb[ANET_MASK_SIZE];
    unsigned index, cur_node, j, size = AS_Size(core);
    unsigned char *status;
    long thrsh = 0;
    fifo f;
	
    if (rnd != 0){
        thrsh = (long)rnd*RAND_MAX;
        srandom(time(0));
    }

    status = (unsigned char*)malloc(size*sizeof(unsigned char));
//    for (unsigned i = 0; i < size; i++) status[i] = WHITE;
    memset(status, WHITE, size*sizeof(unsigned char));

//    for (unsigned i=0; i < size; i++) dest[i] = 0;
    memset(res, 0, size*sizeof(ANET_MASK));
    // initialize FIFO
    fifo_alloc(f, size);

    for (index = 0; index < size; index++){
        if (seeds[index]){
            fifo_insert(f,index);
            status[index] = GREY;
	}
    }

//    while (F.isNotEmpty()){
    while (fifo_notempty(f)){
        cur_node = fifo_extract(&f); //F.Extract()
        if (rnd != 0.0 && random() < thrsh){
//        F.Insert(cur_node);
            fifo_insert(f, cur_node);
            continue;
        }
				
        t_neighbors(AS_Topology(core), cur_node, neighb);
        for (j = 0; j < ANET_MASK_SIZE; j++){
            if (src[cur_node] & (1 << j)) {
                next_node = neighb[j];
                if (status[next_node] == WHITE){
//                    F.Insert(next_node);
                    fifo_insert(f, next_node);
                    status[next_node] = GREY;
                    res[next_node] = 1 << j;
                }
            }
        }
        status[cur_node] = BLACK;
    }

    free(status);
    fifo_free(f);

//    printf("SpanningTree not yet implemented\n");
}

void SPANNING_TREE_STEP(ANET_MASK *res, ANET_MASK *src, Ubyte *seeds, ANET_CORE *core)
{
    register unsigned i, lnk, neighb;
    int nb_link;
    ANET_MASK m;
    unsigned size = AS_Size(core);

  //for (unsigned i=0; i < size; i++)
    //dest[i] &= src[i];

  //Now tmp points to non nil neighbors in sourceg
  for(i = 0; i < size; i++){
    res[i] &= src[i];
    if((m=res[i]) != 0) {
    // get number of neighbors
    nb_link = 0;
    while (m) { nb_link += (m & 1); m = m >> 1; }
    if(nb_link != 1) {
        m = res[i];
	// random number between 0 and linknbr
	/*
        while((rnd=(random()&0x7)) > nb_link) ;
        do{
          for (lnk = 0; lnk < ANET_MASK_SIZE; lnk++)
            if (m & (1 << lnk)) break;
          m &= ~(1 << lnk);
          rnd--;
        }while (rnd>=0);
			*/
        lnk = random() % nb_link;
        neighb = 0;
        while (lnk) { if (m & 1) lnk--; m = m << 1; neighb++; }
        res[i] = (1 << neighb);
      }
    }
  }
 }
