/*
 *  topology.c
 *  Anet
 *
 *  Created by Nicolas Sicard on Tue Sep 03 2002.
 *  Copyright (c) 2002 IEF. All rights reserved.
 *
 *  $Id: topology.c,v 1.1.1.1 2003/09/09 14:41:04 nico Exp $
 *  $Log: topology.c,v $
 *  Revision 1.1.1.1  2003/09/09 14:41:04  nico
 *  Version 0.9
 *
 *  Revision 1.4  2003/07/23 12:39:27  nico
 *  Version fonctionnelle
 *
 *  Revision 1.3  2003/07/07 21:52:25  nico
 *  Modification des keywords CVS
 *
 *  Revision 1.2  2003/07/07 20:23:32  nico
 *  Ajout de mot-cles CVS (auteur, revision, log)
 *
 */

#include "topology.h"

int t_rel[9][ANET_MASK_SIZE];
int **t_rel_mesh = NULL;

void t_rel_init(ANET_TOPOLOGY *topo);

/* allocate memory and initialize topology object */

ANET_TOPOLOGY *newTopologyFromCLI(int argc, char** argv)
{
  if (argc < 2) {
	fprintf(stderr, "Anet:Core:Topo> Missing parameters for topology initialization (width & height) !\n");
	return NULL;
  }
  return newTopology(atoi(argv[0]), atoi(argv[1]));
}

ANET_TOPOLOGY *newTopology(unsigned width, unsigned height)
{
  ANET_TOPOLOGY *topology;

  /* Check topology values */
  if (width < ANET_MIN_WIDTH) {
	fprintf(stderr, "Anet:Core:Topo> Minimum width = 3 !\n");
	return NULL;
  }
  if (width > ANET_MAX_WIDTH) {
	fprintf(stderr, "Anet:Core:Topo> Maximum width = 0xffff !\n");
	return NULL;
  }
  if (height < ANET_MIN_HEIGHT) {
	fprintf(stderr, "Anet:Core:Topo> Minimum height = 3 !\n");
	return NULL;
  }
  if (height > ANET_MAX_HEIGHT) {
	fprintf(stderr, "Anet:Core:Topo> Maximum height = 0xffff !\n");
	return NULL;
  }
  if (width*height > ANET_MAX_SIZE) {
	fprintf(stderr, "** ANET:CORE:TOPOLOGY> Fatal Error : Maximum size = 0xffffff !\n");
	return NULL;
  }
  
  topology = (ANET_TOPOLOGY*)malloc(sizeof(ANET_TOPOLOGY));
  if (topology == NULL) return NULL;
  AT_Width(topology) = width;
  AT_Height(topology) = height;
  AT_Size(topology) = width * height;
  t_rel_init(topology);

  return topology;
}

/* free topology object and information */

void t_rel_init(ANET_TOPOLOGY *topo)
{
	unsigned i;

	t_rel[0][0] = t_rel[1][0] = t_rel[2][0] = topo->size - topo->width;
	t_rel[3][0] = t_rel[4][0] = t_rel[5][0] = t_rel[6][0] = t_rel[7][0] = t_rel[8][0] = -topo->width;
	t_rel[0][1] = t_rel[1][1] = topo->size - topo->width + 1;
	t_rel[2][1] = topo->size - topo->width - topo->width + 1;
	t_rel[3][1] = t_rel[4][1] = t_rel[6][1] = t_rel[7][1] = -topo->width + 1;
	t_rel[5][1] = t_rel[8][1] = 1 - topo->width - topo->width;
	t_rel[0][2] = t_rel[1][2] = t_rel[3][2] = t_rel[4][2] = t_rel[6][2] = t_rel[7][2] = 1;
	t_rel[2][2] = t_rel[5][2] = t_rel[8][2] = 1 - topo->width;
	t_rel[0][3] = t_rel[1][3] = t_rel[3][3] = t_rel[4][3] = topo->width + 1;
	t_rel[2][3] = t_rel[5][3] = 1;
	t_rel[6][3] = t_rel[7][3] = topo->width + 1 - topo->size;
	t_rel[8][3] = -topo->size + 1;
	t_rel[0][4] = t_rel[1][4] = t_rel[2][4] = t_rel[3][4] = t_rel[4][4] = t_rel[5][4] = topo->width;
	t_rel[6][4] = t_rel[7][4] = t_rel[8][4] = topo->width - topo->size;
	t_rel[0][5] = t_rel[3][5] = topo->width + topo->width - 1;
	t_rel[1][5] = t_rel[2][5] = t_rel[4][5] = t_rel[5][5] = topo->width - 1;
	t_rel[6][5] = topo->width + topo->width - topo->size - 1;
	t_rel[7][5] = t_rel[8][5] = topo->width - topo->size - 1;
	t_rel[0][6] = t_rel[3][6] = t_rel[6][6] = topo->width - 1;
	t_rel[1][6] = t_rel[2][6] = t_rel[4][6] = t_rel[5][6] = t_rel[7][6] = t_rel[8][6] = -1;
	t_rel[0][7] = topo->size - 1;
	t_rel[1][7] = t_rel[2][7] = topo->size - topo->width - 1;
	t_rel[3][7] = t_rel[6][7] = -1;
	t_rel[4][7] = t_rel[5][7] = t_rel[7][7] = t_rel[8][7] = -topo->width - 1;

	t_rel_mesh = (int**)malloc(topo->size * sizeof(int*));
	for (i = 0; i < topo->size; i++) {
		if (i == 0) t_rel_mesh[i] = t_rel[0];
		else if (i < topo->width - 1) t_rel_mesh[i] = t_rel[1];
		else if (i == topo->width - 1) t_rel_mesh[i] = t_rel[2];
		else if (i < topo->size - topo->width){
			if ((i % topo->width) == 0) t_rel_mesh[i] = t_rel[3];
			else if (((i+1) % topo->width) == 0)  t_rel_mesh[i] = t_rel[5];
			else t_rel_mesh[i] = t_rel[4];
		}
		else if ((i % topo->width) == 0) t_rel_mesh[i] = t_rel[6];
		else if (i == topo->size-1) t_rel_mesh[i] = t_rel[8];
		else t_rel_mesh[i] = t_rel[7];
	}
}

/* free allocated memory of topology object */
void t_free(ANET_TOPOLOGY *topo)
{
  /* Free memory space allocated for topology structure */
  if (t_rel_mesh != NULL) { free(t_rel_mesh); t_rel_mesh = NULL; }
  if (topo != NULL) free (topo);
  return;
}

/* return topology object buffer size */
unsigned t_buffer_size(ANET_TOPOLOGY *topo)
{
  return (topo->width * topo->height);
}


/*int *t_neighbors_rel(ANET_TOPOLOGY *topo, unsigned i)
{
}*/


/* return an array "neighb" of the neighbors of "i" in topology "topo" */
/* used for random access in a pvar */
void t_neighbors(ANET_TOPOLOGY *topo, unsigned i, unsigned *neighb)
{
    register unsigned size = topo->size;
    register unsigned width = topo->width;
    
    if (i == 0){
        neighb[0] = size - width;
        neighb[1] = neighb[0] + 1;
        neighb[2] = i + 1;
        neighb[3] = i + 1 + width;
        neighb[4] = i + width;
        neighb[5] = width + width - 1;
        neighb[6] = width - 1;
        neighb[7] = size - 1;
        return;
    }
    if (i < width - 1){
        neighb[0] = i + size - width;
        neighb[1] = neighb[0] + 1;
        neighb[2] = i + 1;
        neighb[3] = i + width + 1;
        neighb[4] = i + width;
        neighb[5] = i + width - 1;
        neighb[6] = i - 1;
        neighb[7] = neighb[0] - 1;
        return;
    }
    if (i == width - 1){
        neighb[0] = size - 1;
        neighb[1] = size - width;
        neighb[2] = 0;
        neighb[3] = width;
        neighb[4] = i + width;
        neighb[5] = i + width - 1;
        neighb[6] = i - 1;
        neighb[7] = neighb[0] - 1;
        return;
    }
    if (i < size - width){
        neighb[0] = i - width;
        if (((i+1) % width) == 0){
            neighb[1] = (i + 1) - width - width;
            neighb[2] = (i + 1) - width;
            neighb[3] = i + 1;
        }else{
            neighb[1] = (i + 1) - width;
            neighb[2] = i + 1;
            neighb[3] = i + 1 + width;
        }
        neighb[4] = i + width;
        if ((i % width) == 0){
            neighb[5] = i + width + width - 1;
            neighb[6] = i + width - 1;
            neighb[7] = i - 1;
        }else{
            neighb[5] = i + width - 1;
            neighb[6] = i - 1;
            neighb[7] = i - width - 1;
        }
        return;
    }
    if (i == size - width){
        neighb[0] = i - width;
        neighb[1] = i - width + 1;
        neighb[2] = i + 1;
        neighb[3] = 1;
        neighb[4] = 0;
        neighb[5] = width - 1;
        neighb[6] = size - 1;
        neighb[7] = size - width - 1;
        return;
    }
    if (i < size - 1){
        neighb[0] = i - width;
        neighb[1] = i - width + 1;
        neighb[2] = i + 1;
        neighb[3] = width - (size - i) + 1;
        neighb[4] = width - (size - i);
        neighb[5] = width - (size - i) - 1;
        neighb[6] = i - 1;
        neighb[7] = i - width - 1;
        return;
    }
    neighb[0] = i - width;
    neighb[1] = i - width - width + 1;
    neighb[2] = neighb[0] + 1;
    neighb[3] = 0;
    neighb[4] = width - 1;
    neighb[5] = neighb[4] - 1;
    neighb[6] = i - 1;
    neighb[7] = neighb[0] - 1;
}
