/*
 *  vcontext.h
 *  Anet
 *
 *  Created by Nicolas on Tue Apr 20 2004.
 *  Copyright (c) 2004 IEF. All rights reserved.
 *
 */

#ifndef _VCONTEXT_H_
#define _VCONTEXT_H_

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* maximum number of nested context */
#define VNET_MAX_CTXDEPTH 		16

#define VNET_WOCONTEXT			0x00001000

struct  ANET_VNET;

/* low level structures of context */
/* a virtual context */
typedef struct {
	unsigned char *cond;
	unsigned size;
	int type;
} VNET_CONTEXT;

/* a stck of virtual contexts */
typedef struct {
	unsigned			depth, size;	// depth and default size of contexts
	VNET_CONTEXT		*stack;			// stack (array) of contexts
	VNET_CONTEXT		*current;		// adress of current context
	struct ANET_VNET	*core;			// related virtual net
} VNET_CONTEXT_STACK;

VNET_CONTEXT_STACK *newVContextStack(struct ANET_VNET *vnet);
#define freeVContextStack			vc_free

#define VCONTEXT_WHERE				vc_Where
#define VCONTEXT_ENDWHERE			vc_Endwhere
#define VCONTEXT_ELSEWHERE			vc_Elsewhere

#define VCONTEXT_INIT				vc_init
#define VCONTEXT_FREE				vc_free
#define VCONTEXT_SET_EVERYWHERE		vc_setEverywhere
#define VCONTEXT_SET_NOWHERE 		vc_setNowhere
#define VCONTEXT_PUSH				vc_PushContext
#define VCONTEXT_POP				vc_PopContext
#define VCONTEXT_IS_EVERYWHERE		vc_isEverywhere
#define VCONTEXT_WAS_EVERYWHERE		vc_wasEverywhere
#define VCONTEXT_IS_NOWHERE			vc_isNowhere
#define VCONTEXT_WAS_NOWHERE 		vc_wasNowhere
#define VCONTEXT_IS_PVAR			vc_isPvarContext
#define VCONTEXT_WAS_PVAR			vc_wasPvarContext
#define VCONTEXT_GET_NODES			vc_getContextNodes

//enum { ctxtEverywhere, ctxtNowhere, ctxtPvarBool };

/* context stack initialization and destruction */
void vc_init(VNET_CONTEXT_STACK *ctxs, struct ANET_VNET *vnet);
void vc_free(VNET_CONTEXT_STACK *ctxs);

/* internal context initialisations */
void vc_setEverywhere(VNET_CONTEXT *ctx, unsigned size);
void vc_setNowhere(VNET_CONTEXT *ctx);

/* internal low-level context manipulations and acessors */
#define vc_PushContext(ctxs) { (ctxs)->depth++ ; (ctxs)->current++; }

void vc_PopContext(VNET_CONTEXT_STACK *);

#define vc_isEverywhere(ctx) ((ctx)==NULL || (ctx)->type==ctxtEverywhere)
#define vc_wasEverywhere(ctxs,i) ((ctxs)->stack[(ctxs)->depth+(i)].type == ctxtEverywhere)

#define vc_isPvarContext(ctx) ((ctx) != NULL && (ctx)->type==ctxtPvarBool)
#define vc_wasPvarContext(ctxs,i) ((ctxs)->stack[(ctxs)->depth+(i)].type == ctxtPvarBool)

#define vc_isNowhere(ctx) ((ctx) != NULL && (ctx)->type==ctxtNowhere)
#define vc_wasNowhere(ctxs,i) ((ctxs)->stack[(ctxs)->depth+(i)].type == ctxtNowhere)


#define vc_getContext(ctxs,i) (((ctxs)->depth < VNET_MAX_CTXDEPTH)? ((ctxs)->stack[(ctxs)->depth+(i)]): NULL)
#define vc_getCurrentContext(ctxs) ((ctxs)->current)
#define vc_getPreviousContext(ctxs) ((ctxs)->current - 1)

/* Low level methods (or macros) to get context stack memory space and size */
#define vc_getContextNodes(ctx) (((ctx) != NULL)? (ctx)->cond : NULL)

#define vc_getContextSize(ctx) (((ctx) != NULL)? (ctx)->size : 0)


/* context user fonctions (for C++ methods) */
#define vc_Everywhere(ctxs) \
{ vc_PushContext(ctxs); vc_setEverywhere(vc_getCurrentContext(ctxs), AS_Size((ctxs)->core->compcount)); }

void vc_Where(VNET_CONTEXT_STACK *ctx, unsigned char *cond);
void vc_Endwhere(VNET_CONTEXT_STACK *ctx);
void vc_Elsewhere(VNET_CONTEXT_STACK *ctx);

#endif /* _CONTEXT_H_ */
