/*
 *  vgraph.c
 *  Anet
 *
 *  Created by Nicolas on Tue Sep 09 2003.
 *  Copyright (c) 2003 IEF. All rights reserved.
 *
 */

#include "vgraph.h"
#include "elmtwise.h"
#include "borgefor.h"
#include "mgraph.h"

/** Simplified memory pool to avoid mutexed malloc **/
typedef struct ANET_MPOOL {
    void                *buffer;
    unsigned            size, elmtsize, top;
    struct ANET_MPOOL   *next;
} ANET_MPOOL;

/** Pool preallocation for better efficiency **/
int prealloc_mpool(ANET_MPOOL *mpool, unsigned size)
{
    if (mpool == NULL) return 1;
    if ((mpool->buffer = malloc(size)) == NULL) return 2;
    mpool->size = size;
    mpool->next = NULL;
    mpool->top = 0;
    return 0;
}

/** Liberation du pool memoire **/
void free_mpool(ANET_MPOOL *mpool)
{
    if (mpool == NULL) return;
    if (mpool->buffer) free(mpool->buffer);
    if (mpool->next) { free_mpool(mpool->next); free(mpool->next); }
    return;
}

/** Allocate size bytes on memory pool mpool **/
inline void *anet_malloc(ANET_MPOOL *mpool, unsigned size)
{
	/* Si le pool n'est pas occupé */
    if (mpool->top + size > mpool->size) {
        //printf("Depassement de pool !\n");
        mpool->next = (ANET_MPOOL*)malloc(sizeof(ANET_MPOOL));
        mpool->next->size = (mpool->size < size)? size : mpool->size;
        mpool = mpool->next;
        mpool->next = NULL;
        if ((mpool->buffer = malloc(mpool->size)) == NULL) return NULL;
        mpool->top = size;
        return mpool->buffer;
    } else {
        void *tmp = mpool->buffer + mpool->top;
        mpool->top += size;
        return tmp;
    }
}

/** Free pointer p on memory pool mpool => do nothing **/
inline void anet_free(ANET_MPOOL *mpool, void *p) {/* */}

void vnet_collect_links(ANET_VNET *vnet, ANET_MASK *subnet, ANET_MASK *mask, unsigned *b_compcount, unsigned flag);
void vnet_relabel_components(ANET_VNET *vnet, ANET_MASK *subnet, unsigned *b_compcount, unsigned flag);

/** Initialize virtual net from core (ie base network) **/
void vnet_init(ANET_VNET *vnet, ANET_CORE *core, int label)
{
    unsigned i;

    vnet->compcount = vnet->linkcount = 0;
    vnet->core = core;
    vnet->size = AS_Size(core);
    vnet->component = NULL; // no component for the moment
	if (label) {
		vnet->label = (unsigned*)malloc(vnet->size * sizeof(unsigned));
		memset(vnet->label, 0xff, vnet->size * sizeof(unsigned));
	}else
		vnet->label = NULL;
	/* no links between components */
    for (i = 0; i < ANET_MAX_CONCURRENCY; i++)
        vnet->linksbase[i] = NULL;
	/* context stack */
	vnet->context_stack = NULL;
}

/** Release virtual net data structure **/
void vnet_free(ANET_VNET *vnet, int label)
{
    unsigned    i=0;
    
    if (vnet == NULL) return;
    
	if (vnet->component) { free(vnet->component); vnet->component = NULL; }
	if (label && (vnet->label != NULL)) { free(vnet->label); vnet->label = NULL; }

    while (i < ANET_MAX_CONCURRENCY && vnet->linksbase[i] != NULL) {
        free(vnet->linksbase[i]);
        vnet->linksbase[i] = NULL;
        i++;
    }
	if (vnet->context_stack)
		vc_free(vnet->context_stack);
}

/** Copy vsrc virtual net into vdest **/
void vnet_copy(ANET_VNET *vdest, ANET_VNET *vsrc)
{
	if (vsrc == NULL || vdest == NULL) return;
	vdest->compcount = vsrc->compcount;
	vdest->linkcount = vsrc->linkcount;
	vdest->core = vsrc->core;
	vdest->size = vsrc->size;
	vdest->component = (VNET_COMP*)malloc(vdest->compcount * sizeof(VNET_COMP));
	memcpy(vdest->component, vsrc->component, vdest->compcount * sizeof(VNET_COMP));
	vdest->label = (unsigned*)malloc(vdest->size * sizeof(unsigned));
	memcpy(vdest->label, vsrc->label, vdest->size * sizeof(unsigned));
	/* LINKS ? */
}

/** Allocate and initialize new vnet data structure **/
ANET_VNET *newVnet(ANET_CORE *core)
{
    ANET_VNET *vnet = (ANET_VNET*)malloc(sizeof(ANET_VNET));
    
    vnet_init(vnet, core, 1);
    return vnet;
}

/** Build vnet datas from subnet (mask ??) **/
void vnet_virtualize(ANET_VNET *vnet, ANET_MASK *subnet, ANET_MASK *mask, unsigned flag)
{
    ANET_MASK   *symmetrized;
    unsigned    b_compcount[ANET_MAX_CONCURRENCY], c;
//    NEW_TIME(virt);
    
    if (vnet == NULL || subnet == NULL || mask == NULL) {
        fprintf(stderr, "** ANET:CORE:VGRAPH:VIRTUALIZE: Error : NULL parameter !\n Please check your programm\n");
        exit(1);
    }
    if (flag & VNET_SYMMETRIZE) {
        symmetrized = (ANET_MASK*)malloc(sizeof(ANET_MASK)*vnet->size);
        INVERT_SUBNET(symmetrized, subnet, vnet->core, ANET_SYMMETRIZE);
    } else {
        symmetrized = subnet;
    }

//    START_TIME(virt);
    ELMTWISE_SELF(Address , Uint )(vnet->label, vnet->label, vnet->core, 0);
//    END_TIME(virt);
//    PRINT_TIME(virt, "SelfAddress");
//    START_TIME(virt);
    BORGEFOR_UNSTABLE(Min , Uint)(vnet->label, symmetrized, vnet->core, 0);
//    END_TIME(virt);
//    PRINT_TIME(virt, "Borgefor");
    
//    START_TIME(virt);
    if (AS_BlockCount(vnet->core) < 2)
        vnet_relabel_components(vnet, symmetrized, NULL, flag);
    else
        vnet_relabel_components(vnet, symmetrized, b_compcount, flag);
//    END_TIME(virt);
//    PRINT_TIME(virt, "Relabel");
    //    START_TIME(virt);
    if (AS_BlockCount(vnet->core) < 2)
        vnet_collect_links(vnet, symmetrized, mask, NULL, flag);
    else
        vnet_collect_links(vnet, symmetrized, mask, b_compcount, flag);
	vnet->component[0].iflink = 0;
	for (c = 1; c < vnet->compcount; c++) {
		vnet->component[c].iflink = vnet->component[c-1].iflink + vnet->component[c-1].linkcount;
	}
//    END_TIME(virt);
//    PRINT_TIME(virt, "Liens");
    if (flag & VNET_SYMMETRIZE) {
        free(symmetrized);
        symmetrized = NULL;
    }
}

void vnet_relabel_components(ANET_VNET *vnet, ANET_MASK *subnet, unsigned *b_compcount, unsigned flag)
{
    unsigned    b,v;
    unsigned    *label = vnet->label;
    
    vnet->compcount = 0;
    if (b_compcount == NULL) {
        if (flag & VNET_SINGLETONS) {
            for (v = 0; v < vnet->size; v++) {
                label[v] = (label[v] < v)? label[label[v]] : vnet->compcount++;
            }
        } else {
            for (v = 0; v < vnet->size; v++) {
                if (subnet[v] == 0)
                    label[v] = 0xffffffffu;
                else
                    label[v] = (label[v] < v)? label[label[v]] : vnet->compcount++;
            }
        }
    } else {
        if (flag & VNET_SINGLETONS) {
            for (b = 0; b < AS_BlockCount(vnet->core); b++) {
                for (v = AS_BlockStart(vnet->core, b); v < AS_BlockStart(vnet->core, b+1); v++) {
                    label[v] = (label[v] < v)? label[label[v]] : vnet->compcount++;
                }
                b_compcount[b] = vnet->compcount;
            }
        } else {
            for (b = 0; b < AS_BlockCount(vnet->core); b++) {
                for (v = AS_BlockStart(vnet->core, b); v < AS_BlockStart(vnet->core, b+1); v++) {
                    if (subnet[v] == 0)
                        label[v] = 0xffffffffu;
                    else
                        label[v] = (label[v] < v)? label[label[v]] : vnet->compcount++;
                }
                b_compcount[b] = vnet->compcount;
            }
        }
    }
}

typedef struct VNET_LINK {
    unsigned            component;
    struct VNET_LINK    *next;
} VNET_LINK;

typedef struct {
    VNET_LINK   *link_start;
    unsigned    link_count;
} VNET_LINK_COMP;

typedef struct {
    ANET_VNET       *vnet;
    ANET_MASK       *subnet, *mask;
    unsigned        flag, block, compcount;
    ANET_MPOOL      mpool;
    VNET_LINK_COMP  *lcomps;
} VNET_PARAM;

#define VNET_COLLECT_SUBNET_LINK(cl, cd) { \
    if (lcomps[cl].link_count == 0) { \
        tmp = lcomps[cl].link_start = (VNET_LINK*)anet_malloc(mpool, sizeof(VNET_LINK)); \
            tmp->component = cd; \
                tmp->next = NULL; \
                    lcomps[cl].link_count++; \
    } else { \
        tmp = lcomps[cl].link_start; \
            while (tmp->next != NULL && tmp->component != cd) { \
                tmp = tmp->next; \
            } \
            if (tmp->component != cd) { \
                tmp->next = (VNET_LINK*)anet_malloc(mpool, sizeof(VNET_LINK)); \
                    tmp = tmp->next; \
                        tmp->component = cd; \
                            tmp->next = NULL; \
                                lcomps[cl].link_count++; \
            } \
    } \
}

void *vnet_collect_links_task(VNET_PARAM *p)
{
    unsigned	v, n, begin, end;
    int		*neighbor;
    ANET_VNET   *vnet = p->vnet;
    unsigned	*label = vnet->label;
    ANET_MASK   *mask = p->mask;
    VNET_LINK_COMP	*lcomps;
    ANET_MPOOL  *mpool;
    VNET_LINK   *tmp;
    
    begin = AS_BlockStart(vnet->core,p->block);
    end = AS_BlockStart(vnet->core,p->block+1);
    
    mpool = &(p->mpool);
    lcomps = p->lcomps;
    //printf("t%d : %d composantes\n", p->block, p->compcount);
    memset(lcomps, 0x00, p->compcount * sizeof(VNET_LINK_COMP));
    
    //for (n = 0; n < p->compcount; n++) { lcomps[n].link_start = NULL; lcomps[n].link_count = 0; }
    if (p->flag & VNET_SINGLETONS){
        for (v = begin; v < end; v++) {
            neighbor = t_neighbors_rel(v);
            for (n = 0; n < ANET_MASK_SIZE; n++) {
                if ( (mask[v] & (1 << n)) && (label[v] != label[v+neighbor[n]]) ) {
                    VNET_COLLECT_SUBNET_LINK(label[v], label[v+neighbor[n]])
                }
            }
        }
    } else {
        for (v = begin; v < end; v++) {
            if (label[v] != 0xffffffffu) {
                neighbor = t_neighbors_rel(v);
                for (n = 0; n < ANET_MASK_SIZE; n++) {
                    if ((mask[v] & (1 << n)) && (label[v] != label[v+neighbor[n]]) && (label[v+neighbor[n]] != 0xffffffffu)) {
                        VNET_COLLECT_SUBNET_LINK(label[v], label[v+neighbor[n]])
                    }
                }
            }
        }
    }
    return NULL;
}

typedef struct {
    ANET_VNET   *vnet;
    VNET_PARAM  *vp;
    unsigned    cstart, cend;
    unsigned    block, linkcount;
} MERGE_PARAM;

void vnet_merge_links(MERGE_PARAM *p)
{
    unsigned    lc, c, t, linkcount;
    unsigned    *links;
    VNET_LINK   *ltmp;
    ANET_VNET   *vnet = p->vnet;
    VNET_PARAM  *vp = p->vp;
    
    p->linkcount = 0;
    linkcount = 0;
#define MIN(a,b) (((a)<(b))? (a) : (b))
    for (t = 0; t < AS_BlockCount(vnet->core); t++) {
        for (c = p->cstart; c < MIN(p->cend, vp[t].compcount); c++) {
            linkcount += vp[t].lcomps[c].link_count;
        }
        //        printf("t%d min(c) = %d\n", p->block, MIN(p->cend, vp[t].compcount));
    }
#undef MIN
    memset(vnet->component + p->cstart, 0x00, (p->cend - p->cstart) * sizeof(VNET_COMP));
    vnet->linksbase[p->block] = links = (unsigned*)malloc(sizeof(unsigned)*linkcount);
    if (links == NULL) {
        fprintf(stderr, "** ANET:CORE:VGRAPH:LMERGE> Error : links allocation failed !\n");
        exit(1);
    }
    c = p->cstart;
    //printf("t%d : merge %d %d\n", p->block, p->cstart, p->cend);
    while (c < p->cend) {
        linkcount = 0;
        for (t = 0; t < AS_BlockCount(vnet->core); t++) {
            //printf("t%d: c=%d, t=%d (%d max):\n", p->block, c, t, vp[t].compcount);
            if (c < vp[t].compcount && vp[t].lcomps[c].link_count) {
                //printf("  |-> t%d: c=%d (t=%d): linkcount = %d\n", p->block, c, t, vp[t].lcomps[c].link_count);
                ltmp = vp[t].lcomps[c].link_start;
                while (ltmp != NULL) {
                    lc = 0;
                    while (lc < linkcount && links[lc] != ltmp->component) lc++;
                    //printf("lc = %d, linkcount=%d --\n", lc, linkcount);
                    if (lc == linkcount) {
                        //printf("-<-%d\n", links[lc]);
                        //printf(">-<%d\n", ltmp->component);
                        links[lc] = ltmp->component;
                        //printf("-<<-\n");
                        linkcount++;
                        //printf("-<<<-\n");
                    }
                    ltmp = ltmp->next;
                }
            }
        }
        p->linkcount += linkcount;
        vnet->component[c].links = links;
        vnet->component[c].linkcount = linkcount;
        links += linkcount;
        c++;
    }
    //printf("t%d : end merge\n", p->block);
    vnet->linksbase[p->block] = (unsigned*)realloc(vnet->linksbase[p->block], sizeof(unsigned)*p->linkcount);
    //printf("t%d : p->linkcount = %d\n", p->block,p->linkcount);
}


void vnet_collect_links(ANET_VNET *vnet, ANET_MASK *subnet, ANET_MASK *mask, unsigned *b_compcount, unsigned flag)
{
    VNET_PARAM      param[ANET_MAX_CONCURRENCY];
    MERGE_PARAM     mparam[ANET_MAX_CONCURRENCY];
    pthread_t       task[ANET_MAX_CONCURRENCY];
    unsigned        last_component, status, t, cstart, nbthreads = AS_BlockCount(vnet->core);
//    NEW_TIME(collect);
    
//    START_TIME(collect);
    
    vnet->linkcount = 0;
    
    for(t = 0; t < nbthreads; t++) {
        param[t].vnet = vnet;
        param[t].subnet = subnet;
        param[t].mask = mask;
        param[t].flag = flag;
        param[t].block = t;
        param[t].compcount = (b_compcount != NULL)? (b_compcount[t]) : vnet->compcount;
        param[t].lcomps = (VNET_LINK_COMP*)malloc(sizeof(VNET_LINK_COMP)*(param[t].compcount));
        
        if (prealloc_mpool(&(param[t].mpool), param[t].compcount * 8 * sizeof(VNET_LINK))) {
            fprintf(stderr, "** ANET:CORE:VGRAPH:LINKS> Error : Memory preallocation failed\n");
            exit(1);
        }
        
        status = pthread_create(&(task[t]), NULL, (void*(*)(void*))(vnet_collect_links_task),&(param[t]));
        if (status != 0) fprintf(stderr, "** ANET:CORE:VGRAPH:LINKS> Warning: Thread %d creation failed\n", t);
    }
    
    for(t = 0; t < nbthreads; t++) {
        status = pthread_join(task[t],NULL);
        if (status != 0) fprintf(stderr, "** ANET:CORE:VGRAPH:LINKS> Warning: Thread %d ending failed\n", t);
    }
    
//    END_TIME(collect);
//    PRINT_TIME(collect, "Collecte");
//    START_TIME(collect);
    
    
    last_component = 0;
    vnet->linkcount = 0;
    vnet->component = (VNET_COMP*)malloc(vnet->compcount * sizeof(VNET_COMP));
    
    cstart = 0;
    if (nbthreads > vnet->compcount)
        nbthreads = vnet->compcount;
    for(t = 0; t < nbthreads ; t++) {
        mparam[t].vnet = vnet;
        mparam[t].vp = param;
        mparam[t].block = t;
        mparam[t].cstart = cstart;
        cstart = mparam[t].cend = cstart + (vnet->compcount/nbthreads);
        if (t == nbthreads-1) mparam[t].cend = vnet->compcount;
        //printf("t%d : %d -> %d\n", t, mparam[t].cstart, mparam[t].cend);
        
        status = pthread_create(&(task[t]), NULL, (void*(*)(void*))(vnet_merge_links),&(mparam[t]));
        if (status != 0) fprintf(stderr, "** ANET:CORE:VGRAPH:LMERGE> Warning: Thread %d creation failed\n", t);
    }
    
    vnet->linkcount = 0;
    for(t = 0; t < nbthreads; t++) {
        status = pthread_join(task[t],NULL);
        vnet->linkcount += mparam[t].linkcount;
    }
    for (t = 0; t < AS_BlockCount(vnet->core); t++) {
        free_mpool(&(param[t].mpool));
        free(param[t].lcomps);
    }
    
//    END_TIME(collect);
//    PRINT_TIME(collect, "Fusion");
}

void vnet_print(ANET_VNET *vnet, unsigned width, unsigned flag)
{
    unsigned i;
    
    if (vnet == NULL) return;
    printf("vnet : %d composantes, %d liens\n", vnet->compcount, vnet->linkcount);
    if (flag) {
        for (i = 0; i < vnet->size; i++) {
            if (i % width) printf("%3d ", vnet->label[i]); else printf("\n%3d ", vnet->label[i]);
        }
        printf("\n");
    }
}

#define MASTER_CONTAINER(mc, c) {\
	mc = c; \
	while (container[mc] != mc) mc = container[mc]; \
}

// create a new vnet by component merging
void vnet_virtualize_merge(ANET_VNET *v, Ubyte *mask, ANET_VNET *src)
{
	unsigned			*container, *newindice, *nblinks;
	register unsigned	i, c, cl, nc, mc, l, ccount, lcount;
	ANET_VNET			*vnet = v;

	if (vnet == src) {
		vnet = (ANET_VNET*)malloc(sizeof(ANET_VNET));
		vnet_init(vnet, src->core, 0);
		vnet->label = src->label;
	}
	// components are their own container in the beginning
	lcount = (src->compcount > src->linkcount)? src->compcount : src->linkcount;
	container = (unsigned*)malloc(src->compcount * sizeof(unsigned));
	for (i = 0; i < src->compcount; i++) container[i] = i;
	newindice = (unsigned*)malloc(lcount * sizeof(unsigned));
	memset(newindice, 0xff, lcount * sizeof(unsigned));
	
	// components inclusions scan
	l = 0;
	//printf("vmerge> component inclusions\n");
	//printf("vnet %p : linkcout = %d\n", src, src->linkcount);
	for (c = 0; c < src->compcount; c++) {
		MASTER_CONTAINER(mc, c);			// mc <- master component of c
		for (cl = 0; cl < src->component[c].linkcount; cl++, l++) {
			nc = src->component[c].links[cl];   // neighbor component of c
			//printf("%d<-%d;", c, nc);
			if (mask[l/8] & (1 << (l%8))) {
				//printf("%d<-%d;", c, nc);
				if (container[nc] == nc && mc != nc) {
					container[nc] = mc;
				}
			}
		}
	}

	//printf("l = %d\n", l);
	for (c = 0; c < src->compcount; c++) {
		MASTER_CONTAINER(mc, c);			// mc <- master component of c
		container[c] = mc;
	}
	
	//components counting and relabelling
	ccount = 0;
	for (c = 0; c < src->compcount; c++) {
		if (container[c] == c) {
			newindice[c] = ccount++;
			/*printf(" %d:%2d,",c, newindice[c]);
		} else {
			printf(" %d:[%d],",c, container[c]);*/
		}
	}
	//printf("\n");
	
	// nodes relabelling
	for (c = 0; c < src->size; c++) {
		vnet->label[c] = newindice[container[src->label[c]]];
	}
	//printf("%d nouvelles composantes\n", ccount);
	
	vnet->compcount = ccount;
	vnet->component = (VNET_COMP*)malloc(sizeof(VNET_COMP)*ccount);
	//memset(vnet->component, 0x00, sizeof(VNET_COMP)*ccount);
	
	if (ccount > 1) {
		
		nblinks = (unsigned*)malloc(sizeof(unsigned) * vnet->compcount);
		memset(nblinks, 0x00, sizeof(unsigned) * vnet->compcount);
		for (c = 0; c < src->compcount; c++) {
			container[c] = newindice[container[c]]; //container contains new indice of component
			nblinks[container[c]] += src->component[c].linkcount; //nblinks contains total link number
		}
		
		memset(newindice, 0xff, lcount * sizeof(unsigned));
		vnet->component[0].links = newindice;
		vnet->component[0].linkcount = 0;
		//for (c = 0; c < vnet->compcount; c++) printf("nl[%d]=%d,",c, nblinks[c]);
		for (c = 1; c < vnet->compcount; c++) {
			vnet->component[c].links = vnet->component[c-1].links + nblinks[c-1];
			vnet->component[c].linkcount = 0;
		}
		//printf("---\n");

		//printf("link collecting...\n");
		//links collecting
		lcount = 0;
		for (c = 0; c < src->compcount; c++) {
			mc = container[c];
			for (cl = 0; cl < src->component[c].linkcount; cl++) {
				l = 0;
				while ((l < vnet->component[mc].linkcount) &&
					   (container[src->component[c].links[cl]] != vnet->component[mc].links[l])) l++;
				if (l == vnet->component[mc].linkcount && container[src->component[c].links[cl]] != mc) {
					vnet->component[mc].links[l] = container[src->component[c].links[cl]];
					(vnet->component[mc].linkcount)++;
					(vnet->linkcount)++;
				}
			}
		}
		vnet->linksbase[0] = newindice;
		//printf("\n");
		
		// component link indexing
		vnet->component[0].iflink = 0;
		for (c = 1; c < vnet->compcount; c++) {
			vnet->component[c].iflink = vnet->component[c-1].iflink + vnet->component[c-1].linkcount;
		}
		
		free(nblinks);
		
	} else {
		
		vnet->component[0].links = NULL;
		vnet->component[0].linkcount = 0;
		memset(vnet->label, 0x00, vnet->size * sizeof(unsigned));
		vnet->linksbase[0] = NULL;
	}
	
	
	if (v == src) {
		vnet_free(src, 0);
		src->compcount = vnet->compcount;
		src->linkcount = vnet->linkcount;
		src->linksbase[0] = vnet->linksbase[0];
		src->component = vnet->component;
		free(vnet);
	}
	free(container);
}

/**** VNET REDUCTION ****/

/* Initialize an allocated rnet data structure */
void vrnet_init(ANET_VRNET *vrnet, ANET_VNET *vnet)
{
    vrnet->compcount = 0;			/* no component */
    vrnet->vnet = vnet;				/* base virtual graph */
    vrnet->size = vnet->compcount;   /* virtual graphe size */

	/* preallocate as many components as nodes in base virtual graph (ie maximum component count) */
    vrnet->scc = (VRNET_SCC*)malloc(vrnet->size * sizeof(VRNET_SCC));
	/* initialize components to NULL */
    memset(vrnet->scc, 0x00, vrnet->size * sizeof(VRNET_SCC));
	/* allocate and initialize to MAX_INT nodes component labels (0 is a valid index) */
    vrnet->label = (unsigned*)malloc(vrnet->size * sizeof(unsigned));
    memset(vrnet->label, 0xff, vrnet->size * sizeof(unsigned));
}

/* Release an allocated and build rnet */
void vrnet_free(ANET_VRNET *vrnet)
{
    VRNET_LINK  *link;  /* to go through predecessors list */
    unsigned	c;		/* component iterator */
	
    if (vrnet == NULL) return;
	
    /* components have been relabeled */
    for (c = 0; c < vrnet->compcount; c++)
        while (vrnet->scc[c].links != NULL) {
            link = vrnet->scc[c].links;
            vrnet->scc[c].links = link->next;
            free(link);
        }
	/* release scc array and label list */
	if (vrnet->scc) free(vrnet->scc);
    if (vrnet->label) free(vrnet->label);
}

/* Return a preallocated and initialized vrnet data structure address */
ANET_VRNET *newVRnet(ANET_VNET *vnet)
{
    ANET_VRNET *vrnet = (ANET_VRNET*)malloc(sizeof(ANET_VRNET));	/* allocate */
    
    vrnet_init(vrnet, vnet);  /* initialize */
    return vrnet;
}

#include "stack.h"

/* Macro for parameter*/
#define FIRSTBLOCK(vrp) ((vrp)->firstblock)
#define BLOCKCOUNT(vrp) ((vrp)->blockcount)
#define DFSNUM(vrp, v) ((vrp)->dfsnum[v])

#define IS_UNPROCESSED(vrp,v) ((vrp)->dfsnum[v] == 0xffffffffu)

#define ROOT		0x01
#define UNFINISHED	0x02

#define IS_UNFINISHED(vrp, v) ((vrp)->is[v] & UNFINISHED)
#define IS_ROOT(vrp, v) ((vrp)->is[v] & ROOT)

/* Data parameter for threaded functions */
typedef struct {
    ANET_VRNET		*vrnet;					/* used reduced net data structure */
    Ubyte			*vsubnet;				/* original subnet */
    unsigned		compcount;				/* number of strong connected components */
    unsigned		*dfsnum, count1;		/* for DFS use */
    stack			unfinished, root;		/* for DFS use */
    unsigned char	*is;					/* for DFS use : flags "is UNFINISHED" or "is ROOT" */	
} VRNET_PARAM;

void vrnet_explore_node(VRNET_PARAM *, unsigned);
void vrnet_collect_links(VRNET_PARAM *);
/* Recursive vnet reduction*/
void vrnet_reduce(ANET_VRNET *vrnet, Ubyte *vsubnet)
{
    unsigned		*dfsnum, v;			/* dfs numbers and parent scc indices arrays */
    unsigned char	*is;				/* is flags array */
    VRNET_PARAM		vrp;				/* vrnet data parameter */

    /* allocate dfsnum, pscc and is arrays */
    dfsnum = (unsigned*)malloc(vrnet->size*sizeof(unsigned));
    is = (unsigned char*)malloc(vrnet->size*sizeof(unsigned char));
    memset(is, 0x00, vrnet->size*sizeof(unsigned char));
    memset(dfsnum, 0xff, vrnet->size*sizeof(unsigned));
	/* allocate and initialize stacks for dfs (tarjan algorithm) */
    stack_alloc(vrp.unfinished, vrnet->size+1);
    stack_init(vrp.unfinished);
    stack_alloc(vrp.root, vrnet->size+1);
    stack_init(vrp.root);
	
	/* initialize parameter fields */
    vrp.vrnet = vrnet; vrp.vsubnet = vsubnet;
    vrp.count1 = vrp.compcount = 0; vrp.dfsnum = dfsnum; vrp.is = is;

	/* process nodes */
	for (v = 0; v < vrnet->size; v++) {
		if (IS_UNPROCESSED((&vrp),v)) vrnet_explore_node(&vrp, v);
	}
	
	/* memorize number of components */
	vrnet->compcount = vrp.compcount;

	/* collect inter-component links */
	vrnet_collect_links(&vrp);
	/* release potentially *huge* wasted memory in scc array */
	vrnet->scc = (VRNET_SCC*)realloc(vrnet->scc, vrnet->compcount*sizeof(VRNET_SCC));
	/* free DFS data structures */
	free(dfsnum); free(is);
}

// IL RESTE A TRAITER LES LIENS INTER-COMPOSANTES !!
/* DFS node exploration for scc detection (Tarjan) */
void vrnet_explore_node(VRNET_PARAM *rp, unsigned v)
{
    unsigned 	*neighbor;
    unsigned	n, w, l;
	VNET_COMP	*component = rp->vrnet->vnet->component;
	
	/* attributes v a dfs number */
    DFSNUM(rp, v) = ++(rp->count1);
	/* in the beginning v is unfinished, and a component root */
    stack_push(rp->unfinished, v);
    rp->is[v] |= UNFINISHED;
    stack_push(rp->root, v);
    rp->is[v] |= ROOT;
	
	/* get v neighbors list */
    neighbor = component[v].links;
	l = component[v].iflink;
	/* for each nth neighbor */
    for (n = 0; n < component[v].linkcount; n++) {
		/* if its link with v is activated */
        if (rp->vsubnet[l/8] & (1 << (l%8))) {
			/* get its absolute indice w */
            w = neighbor[n];
			/* if w is also unprocessed resume exploration through it */
            if (IS_UNPROCESSED(rp, w)) {
                vrnet_explore_node(rp,w);
            } else { /* we've already seen w */
				/* if it's still active */
                if (IS_UNFINISHED(rp, w)) {
					/* all nodes explores from w belong to the same component ! */
                    while (DFSNUM(rp, stack_head(rp->root)) > DFSNUM(rp, w)) {
						/* they can't be root (extract them from the root stack) */
                        rp->is[stack_pop(rp->root)] &= ~ROOT;
					}
                }
            }
        }
		l++;
    }
	
	/* all that could be explored from v is explored and processed */
	/* if v is still in the root stack... */
    if (v == stack_head(rp->root)) {
		/* ... then that's because it's the root of a new strong connected component ! */
		/* but not anymore !*/
        rp->is[stack_pop(rp->root)] &= ~ROOT;
		/* let's pop unfinished nodes w until we reach down to v */
        do {
            w = stack_pop(rp->unfinished);
			/* now w is finished... */
            rp->is[w] &= ~UNFINISHED;
			/* ...and it belongs to the component rp->compcount*/
            rp->vrnet->label[w] = rp->compcount;
        } while (w != v);
		/* for now the component is its own englobing component */
        //rp->pscc[rp->compcount] = rp->compcount;
		/* ok, don't forget we've got one more component ! */
        rp->compcount++;
    }
}

//static unsigned g_linkcount = 0;

/* Macro definition for inserting component links with duplicate removal */
#define COLLECT_VSUBNET_LINK(cl, cd) { \
    tmp = scc[cl].links; \
	if (tmp == NULL) { \
		tmp = scc[cl].links = (VRNET_LINK*)malloc(sizeof(VRNET_LINK)); \
		tmp->scc = cd; \
		tmp->next = NULL; \
	} else { \
		while ((tmp->scc != cd) && (tmp->next != NULL)) { tmp = tmp->next; } \
		if (tmp->scc != cd) { \
			tmp = scc[cl].links; \
			scc[cl].links = (VRNET_LINK*)malloc(sizeof(VRNET_LINK)); \
			(scc[cl].links)->scc = cd; \
			(scc[cl].links)->next = tmp; \
		} \
	} \
} \

void vrnet_collect_links(VRNET_PARAM *vrp)
{
	register unsigned	v, w, l, n;
	VNET_COMP	*component = vrp->vrnet->vnet->component;	
	unsigned	*label = vrp->vrnet->vnet->label;
	VRNET_SCC	*scc = vrp->vrnet->scc;
    VRNET_LINK	*tmp;
	
	l = 0;
	//g_linkcount = 0;
	for(v = 0; v < vrp->vrnet->size; v++) {
		for (n = 0; n < component[v].linkcount; n++,l++) {
			w = component[v].links[n];
			/* if w connected to v and scc[v] != scc[w] */
			if (vrp->vsubnet[l/8] & (1 << (l%8)) && (label[v] != label[component[v].links[n]])) {
				COLLECT_VSUBNET_LINK(label[v], label[component[v].links[n]]);
			}
		}
	}
	//printf("%d liens au total\n", g_linkcount);
}
