/*
 *  vgraph.h
 *  Anet
 *
 *  Created by Nicolas on Tue Sep 09 2003.
 *  Copyright (c) 2003 IEF. All rights reserved.
 *
 */

#ifndef _VGRAPH_H_
#define _VGRAPH_H_

#include "core.h"
#include "vcontext.h"

typedef struct {
    unsigned	*links;
    unsigned    linkcount, iflink;
} VNET_COMP;

typedef struct ANET_VNET {
    ANET_CORE	*core;
    unsigned	size, compcount;
    unsigned    linkcount;
    VNET_COMP	*component;
    unsigned	*label;
    unsigned    *linksbase[ANET_MAX_CONCURRENCY];
	VNET_CONTEXT_STACK  *context_stack;
} ANET_VNET;

#define     VNET_SINGLETONS     1U
#define     VNET_SYMMETRIZE     2U

ANET_VNET *newVnet(ANET_CORE *core);
void vnet_free(ANET_VNET *vnet, int label);
void vnet_virtualize(ANET_VNET *vnet, ANET_MASK *subnet, ANET_MASK *mask, unsigned flag);
void vnet_copy(ANET_VNET *vsrc, ANET_VNET *vdest);
void vnet_print(ANET_VNET *vnet, unsigned width, unsigned flag);

void vnet_virtualize_merge(ANET_VNET *vnet, Ubyte *mask, ANET_VNET *src);

/** VSUBNET REDUCTION **/
/* Link cell for list of component predecessors */
typedef struct VRNET_LINK {
    unsigned			scc;						/* index of predecessor scc */
    struct VRNET_LINK	*next;						/* next link cell in linked list */
} VRNET_LINK;

/* Strong Connected Component data structure */
typedef struct {
    VRNET_LINK	*links;								/* Linked list of scc dependancies */
} VRNET_SCC;

/* Reduced graph data structure */
typedef struct {
    ANET_VNET   *vnet;								/* Relative vnet */
    unsigned	size;								/* Size of relative vnet */
	unsigned	compcount;							/* Number of strong connected components */
    VRNET_SCC	*scc;								/* Array of strong connected components */
    unsigned	*label;								/* Array of node's component index in "scc" array*/
} ANET_VRNET;

void vrnet_init(ANET_VRNET *vrnet, ANET_VNET *vnet);
/* Release an allocated and build rnet */
void vrnet_free(ANET_VRNET *vrnet);
/* Return a preallocated and initialized vrnet data structure address */
ANET_VRNET *newVRnet(ANET_VNET *vnet);
/* Recursive vnet reduction*/
void vrnet_reduce(ANET_VRNET *vrnet, Ubyte *vsubnet);

#endif /* _VGRAPH_H_ */
