/*
 *  anetcore.cpp
 *  Anet
 *
 *  Created by Nicolas Sicard on Tue Jan 21 2003.
 *  Copyright (c) 2003 IEF. All rights reserved.
 *
 * $Author: nico $
 * $Revision: 1.1.1.1 $
 * $Log: anetcore.cc,v $
 * Revision 1.1.1.1  2003/09/09 14:41:04  nico
 * Version 0.9
 *
 * Revision 1.6  2003/07/17 09:44:45  nico
 * Ajout d'une methode Initialisation a partir d'un nom de fichier et du nombre de threads. Transfert de ImageFileSize depuis pvar.th dans anetcore.h
 *
 * Revision 1.5  2003/07/08 13:56:24  nico
 * Ajout des méthodes Width() et Height()
 *
 * Revision 1.4  2003/07/07 21:48:24  nico
 * Ajout des methodes de classe GetConcurrency et SetConcurrency
 *
 * Revision 1.3  2003/07/07 21:31:35  nico
 * Retrait des mot-cles CVS pour cause d'incompatibilite avec perl tpp !
 *
 */

#include "anetcore.h"
#include "pvarbool.h"

/* Data structure for Anet library */
ANET_CORE*	Anet::core = NULL;

/* Library initialization from width, height and concurrency as command line args */
int Anet::Initialize(int argc, char **argv)
{
  if (argc < 3 || argv == NULL) return 1;
  return Initialize(atoi(argv[0]), atoi(argv[1]), atoi(argv[2]));
}

/* Library initialization from width, height and concurrency integers */
/* concurrency usually refers to the # of threads */
int Anet::Initialize(unsigned width, unsigned height, int concurrency)
{
  core = sharedCore(newTopology(width, height), concurrency);
  if (core == NULL) return 1;		/* Initialization Error */
  return 0;
}

/* Library initialization from width, height and concurrency integers */
/* concurrency usually refers to the # of threads */
int Anet::Initialize(const char *fname, int concurrency)
{
    unsigned width, height;
    
    if (Anet::ImageFileSize(fname, &width, &height)) return 1;
    core = sharedCore(newTopology(width, height), concurrency);
    if (core == NULL) return 1;		/* Initialization Error */
    return 0;
}


/* release resources needed by Anet */
void Anet::End()
{
  closeCore(core);
}

/* Set context status */
void Anet::Where(const PvarBool& cond) { CONTEXT_WHERE(AS_ContextStack(core), cond.getBuffer()); }

int Anet::SetConcurrency(unsigned concurrency)
{
    if (core == NULL) return 1;
    setCoreConcurrency(core, concurrency);
    return 0;
}

unsigned int Anet::Concurrency()
{
    if (core == NULL) return 0;
    return getCoreConcurrency(core);
}


/* Command line parameters parser */
void Anet::CLIParser(int argc, char *const *argv, const char* optlist,
		     int (*optproc)(int, const char*)){
  int index = 1;
  while (1){
    if (index >= argc) return;
    if (argv[index][0] == '-' && argv[index][1] != '-'){
      for (unsigned i = 0; i < strlen(optlist); i++){
	if (optlist[i] == argv[index][1]){
	  if (i < strlen(optlist)-2 && optlist[i+1] == ':' && optlist[i+2] == ':'){
	    //require an optional argument
	    if (index < argc-1 && argv[index+1][0] != '-'){
	      (*optproc)((int)optlist[i],argv[index+1]);
	      index++;
	    }else
	      (*optproc)((int)optlist[i],(char*)0L);
	  }
	  else if (i < strlen(optlist)-1 && optlist[i+1] == ':'){
	    //require an argument
	    if (index < argc-1 && argv[index+1][0] != '-'){
	      (*optproc)((int)optlist[i],argv[index+1]);
	      index++;
	    }else{
	      cerr << "Erreur de ligne de commande :" << endl;
	      cerr << "  l'option -" << optlist[i] << " exige un argument." << endl;
	      cerr << "Programme arrete." << endl;
	      End();
	      exit(1);
	    }
	  }
	  else{
	    //require no argument
	    (*optproc)((int)optlist[i],(char*)0L);
	    index++;
	  }
	  break;
	}
      }
    }
    else { index++; }
  }
}
