/*
 *  anetcore.h
 *  Anet
 *
 *  Created by Nicolas Sicard on Tue Jan 21 2003.
 *  Copyright (c) 2003 IEF. All rights reserved.
 *
 * $Author: nico $
 * $Revision: 1.1.1.1 $
 * $Log: anetcore.h,v $
 * Revision 1.1.1.1  2003/09/09 14:41:04  nico
 * Version 0.9
 *
 * Revision 1.5  2003/07/17 09:44:45  nico
 * Ajout d'une methode Initialisation a partir d'un nom de fichier et du nombre de threads. Transfert de ImageFileSize depuis pvar.th dans anetcore.h
 *
 * Revision 1.4  2003/07/08 13:56:24  nico
 * Ajout des méthodes Width() et Height()
 *
 * Revision 1.3  2003/07/07 21:48:24  nico
 * Ajout des methodes de classe GetConcurrency et SetConcurrency
 *
 * Revision 1.2  2003/07/07 19:08:22  nico
 * Ajout de mot-cles CVS (auteur, revision, log)
 *
 */

#ifndef _ANETCORE_H_
#define _ANETCORE_H_

#include <stream.h>

extern "C"{
#include "core.h"
#include "files.h"
}

/* Predeclaration of PvarBool for Where(PvarBool&) methode*/
class PvarBool;

/* Static class for centralized library informations and global fonctions */
class Anet {
protected:
  static ANET_CORE*	core;

/* Library initialization */
public :
  static int Initialize(int argc, char **argv);		//from command line ("size size concurrency")
  static int Initialize(unsigned width, unsigned height, int nbtask);	//from values
  static int Initialize(const char *fname, int concurrency);		//from image file and concurrency value

  static void End();
  
#ifdef ANET_VERS
  static const char *Version() { return ANET_VERS; }
#endif /* ANET_VERS */
  static void CLIParser(int argc, char *const *argv, const char* optlist,
                        int (*optproc)(int, const char*));
  static int SetConcurrency(unsigned);
  static unsigned int Concurrency();
  static int ImageFileSize(const char *filename, unsigned int *width, unsigned int *height) {
    return READ_PVARFILE_SIZE(filename, width, height);
  }


/* Topology informations */
public:
  unsigned Size() const { return AS_Size(core); }
  unsigned Width() const { return AS_Width(core); }
  unsigned Height() const { return AS_Height(core); }
  unsigned Degree() const { return ANET_MASK_SIZE; }
  static ANET_CORE *Core() { return core; }
  
/* Context related methods */
public:
  /* Check context status */
  static int IsEverywhere(int i=0) { return CONTEXT_IS_EVERYWHERE(AS_Context(core)); }
  static int IsPvarContext(int i=0) { return CONTEXT_IS_PVAR(AS_Context(core)); }
  static int IsNowhere(int i=0) { return CONTEXT_IS_NOWHERE(AS_Context(core)); }
  /* Set context status */
  static void Everywhere() { CONTEXT_SET_EVERYWHERE(AS_Context(core), AS_Size(core)); }
  static void Where(const PvarBool& cond);
  static void Endwhere() { CONTEXT_ENDWHERE(AS_ContextStack(core)); }
  static void Elsewhere() { CONTEXT_ELSEWHERE(AS_ContextStack(core)); }
  /*
protected:
  static void PushContext() { c_PushContext(&context); }
  static void PopContext() { c_PopContext(&context); }
  // Low level methods to get context stack memory space and size
  static unsigned *getContext() { return c_getContext(&context); }
  static unsigned getContextSize() { return c_getContextSize(&context); }
     */
};

#endif /* _ANETCORE_H_ */
