/*
 *  subnet.cc
 *  Anet
 *
 *  Created by Nicolas Sicard on Jul 07 2003.
 *  Copyright (c) 2003 IEF. All rights reserved.
 *
 * $Id: subnet.cc,v 1.1.1.1 2003/09/09 14:41:04 nico Exp $
 * $Log: subnet.cc,v $
 * Revision 1.1.1.1  2003/09/09 14:41:04  nico
 * Version 0.9
 *
 * Revision 1.5  2003/07/25 20:28:00  nico
 * Renommage de invert.h en mgraph.h
 *
 * Revision 1.4  2003/07/20 21:07:21  nico
 * Correction diverses apres compilation -Wall, virtualisation et reduction pour vnet (vnet.cc)
 *
 * Revision 1.2  2003/07/07 19:05:38  nico
 * Ajout de mot-cles CVS (auteur, revision, log)
 */

#include "subnet.h"
#include "pvarbool.h"

/* Anet Core C files */
extern "C"{
#include "mgraph.h"
#include "spanning.h"
}

/***************************************************************/
/*                        Implementation                       */
/***************************************************************/
/*
ostream& operator<<(ostream& out, const Subnet& p){
  if (p.Net()->Type() == MESH){
    Mesh *mesh = (Mesh*)p.Net();
    out << mesh->XSize() << " " << mesh->YSize() << endl;
  }else{
    out << p.Size() << endl;
  }
  for(unsigned i = 0; i < p.Size(); i++)
    out << (unsigned)p.pvBuffer[i] << " ";
  out << "Subnet:<< not yet implemented" << endl;
  return out;
}
*/

Subnet::Subnet(const Subnet& s) : PvarMask(s)
{
  invert = NULL;	// not used yet
	rnet = NULL;
}

Subnet::Subnet(const PvarMask& p) : PvarMask(p)
{
  invert = NULL;	// not used yet
	rnet = NULL;
}

Subnet::Subnet(const ANET_MASK& m) : PvarMask(m)
{
  invert = NULL;	// not used yet
	rnet = NULL;
}

Subnet::~Subnet(){
  if (invert != NULL) delete invert;
	if (rnet != NULL) rnet_free(rnet);
}

Subnet& Subnet::operator=(const Subnet& s)
{
  PvarMask::operator=(s);
	FlushRnet();
  return *this;
}

Subnet& Subnet::operator=(const PvarMask& a)
{
  PvarMask::operator=(a);
	FlushRnet();
  return *this;
}

Subnet& Subnet::operator=(const ANET_MASK& m)
{
  PvarMask::operator=(m);
	FlushRnet();
  return *this;
}

Subnet& Subnet::Union(const Subnet& s)
{
	FlushRnet();
  return ((*this) |= s);
}

Subnet& Subnet::Intersection(const Subnet& s)
{
	FlushRnet();
  return ((*this) &= s);
}

Subnet& Subnet::operator&=(const Subnet& s)
{
  PvarMask::operator&=(s);
	FlushRnet();
  return *this;
}

Subnet& Subnet::operator&=(const ANET_MASK& m)
{
  PvarMask::operator&=(m);
	FlushRnet();
  return *this;
}

Subnet& Subnet::operator|=(const Subnet& s)
{
  PvarMask::operator|=(s);
	FlushRnet();
  return *this;
}

Subnet& Subnet::operator|=(const ANET_MASK& m)
{
  PvarMask::operator|=(m);
	FlushRnet();
  return *this;
}

Subnet Subnet::operator&(const Subnet& s) const
{
  Subnet tmp(*this);
  tmp &= s;
  return tmp;
}

Subnet Subnet::operator&(const ANET_MASK& m) const
{
  Subnet tmp(*this);
  tmp &= m;
  return tmp;
}

Subnet Subnet::operator|(const Subnet& s) const
{
  Subnet tmp(*this);
  tmp |= s;
  return tmp;
}

Subnet Subnet::operator|(const ANET_MASK& m) const
{
  Subnet tmp(*this);
  tmp |= m;
  return tmp;
}

void Subnet::Invert()
{
  INVERT_SUBNET(pvBuffer, pvBuffer, core, ANET_INVERT);
	FlushRnet();
}

void Subnet::Symmetrize()
{
  INVERT_SUBNET(pvBuffer, pvBuffer, core, ANET_SYMETRIZE);
	FlushRnet();
}

void Subnet::Mask(int m)
{
  MASK_SUBNET(pvBuffer, core, m);
	FlushRnet();
}

void Subnet::SpanningTree(const Subnet& s, const PvarBool& p, float rnd)
{
  SPANNING_TREE(pvBuffer, s.getBuffer(), p.getBuffer(), rnd, core);
	FlushRnet();
}

void Subnet::SpanningTreeStep(const Subnet& s, const PvarBool& p)
{
  SPANNING_TREE_STEP(pvBuffer, s.getBuffer(), p.getBuffer(), core);
	FlushRnet();
}
