/*
 *  subnet.h
 *  Anet
 *
 *  Created by Nicolas Sicard on Jul 07 2003.
 *  Copyright (c) 2003 IEF. All rights reserved.
 *
 * $Id: subnet.h,v 1.1.1.1 2003/09/09 14:41:04 nico Exp $
 * $Log: subnet.h,v $
 * Revision 1.1.1.1  2003/09/09 14:41:04  nico
 * Version 0.9
 *
 * Revision 1.3  2003/07/20 21:07:21  nico
 * Correction diverses apres compilation -Wall, virtualisation et reduction pour vnet (vnet.cc)
 *
 * Revision 1.2  2003/07/07 19:05:38  nico
 * Ajout de mot-cles CVS (auteur, revision, log)
 */

#ifndef _SUBNET_H_
#define _SUBNET_H_

//#include <fstream.h>
#include "pvar.h"

#define PvarMask PvarUbyte

/***************************************************************/
/*                             Subnet                          */
/***************************************************************/

class Subnet : public PvarMask{
public:
	ANET_RNET *rnet;
protected:
	ANET_MASK *invert;

public:
	Subnet() : PvarMask() { invert = NULL; rnet = NULL; }
	Subnet(const Subnet& s);
	Subnet(const PvarMask& pm);
	Subnet(const ANET_MASK& m);
	~Subnet();

	Subnet& operator=(const Subnet&);
	Subnet& operator=(const PvarMask&);
	Subnet& operator=(const ANET_MASK&);

  /* Subnet Union & Intersection*/
  Subnet& Union(const Subnet&);
  Subnet& Intersection(const Subnet&);

  ANET_MASK operator()(unsigned i) const { return pvBuffer[i]; };

  void Invert();
  void Symmetrize();

  void SpanningTree(const Subnet& s, const PvarBool&, float rnd);
  void SpanningTreeStep(const Subnet& s, const PvarBool&);

  void Mask(int m);

  friend Subnet operator~(const Subnet& s){
    Subnet tmp;
    for (unsigned i = 0; i < tmp.Size(); i++)
      tmp.pvBuffer[i] = ~s.pvBuffer[i];
    return tmp;
  }

  //some useful aoperators
  Subnet& operator&=(const Subnet&);
  Subnet& operator|=(const Subnet&);
  Subnet& operator&=(const ANET_MASK&);
  Subnet& operator|=(const ANET_MASK&);
  Subnet operator&(const Subnet&) const;
  Subnet operator|(const Subnet&) const;
  Subnet operator&(const ANET_MASK&) const;
  Subnet operator|(const ANET_MASK&) const;
/*
  void WriteASN(const char *name){
    if (name == (const char*)0L) return;
    ofstream fout(name);
    fout << *this << endl;
    fout.close();
  }
*/
  ANET_MASK *InvertedNet();
  void FlushRnet() { if (rnet != NULL) rnet_free(rnet); rnet = NULL; }
};


#endif
