/*
 *  vnet.cc
 *  Anet
 *
 *  Created by Nicolas on Thu Jun 26 2003.
 *  Copyright (c) 2003 IEF. All rights reserved.
 *
 * $Author: nico $
 * $Revision: 1.1.1.1 $
 * $Log: vnet.cc,v $
 * Revision 1.1.1.1  2003/09/09 14:41:04  nico
 * Version 0.9
 *
 * Revision 1.6  2003/07/25 16:20:42  nico
 * Retour dans le repertoire Source/Object
 *
 * Revision 1.1  2003/07/25 13:09:58  nico
 * Deplacement des fichier dans l'arborescence Virtual => bibliotheque a part
 *
 * Revision 1.5  2003/07/23 12:39:37  nico
 * Version fonctionnelle
 *
 * Revision 1.4  2003/07/21 13:35:55  nico
 * Ajout d'un flag (impression des valeurs ou non) pour la methode print
 *
 * Revision 1.3  2003/07/20 21:07:21  nico
 * Correction diverses apres compilation -Wall, virtualisation et reduction pour vnet (vnet.cc)
 *
 * Revision 1.2  2003/07/07 19:08:22  nico
 * Ajout de mot-cles CVS (auteur, revision, log)
 *
 */

#include "vnet.h"
#include "vsubnet.h"
#include "vpvarbool.h"

/* Vnet class methods */
Vnet*  Vnet::imp_vnet = NULL;

void Vnet::With(Vnet& v)
{
	if (imp_vnet != NULL)
		fprintf(stderr, "VNET> Warning : WITH superposition not (yet) allowed\n");
	imp_vnet = &v;
}
void Vnet::EndWith() { imp_vnet = NULL; }

/* Vnet instance methods */
Vnet::Vnet()
{
    vnet = newVnet(core);
}

Vnet::Vnet(const Vnet& v)
{
    vnet = newVnet(core);
    vnet_copy(vnet, v.vnet);
	// initialize internal context stack
	vnet->context_stack = newVContextStack(vnet);
}

Vnet::Vnet(Subnet& s, Subnet& mask, int flag)
{
    vnet = newVnet(core);
	vnet_virtualize(vnet, s.getBuffer(), mask.getBuffer(), flag);
	// initialize internal context stack
	vnet->context_stack = newVContextStack(vnet);
}
/*
Vnet::Vnet(Subnet& s, unsigned mask_id)
{
	Subnet  mesh(0xff);
	
	mesh.Mask(mask_id);
    vnet = newVnet(core);
	vnet_virtualize(vnet, s.getBuffer(), mesh.getBuffer(), 0);
}
*/


Vnet::~Vnet()
{
    vnet_free(vnet, 1);
}

void Vnet::Virtualize(Subnet& s, Subnet& mask, int flag)
{
	vnet_virtualize(vnet, s.getBuffer(), mask.getBuffer(), flag);
	// (re)initialize internal context stack
	if (vnet->context_stack != NULL)
		vc_free(vnet->context_stack);
	vnet->context_stack = newVContextStack(vnet);
}
/*
void Vnet::Virtualize(Subnet& s, unsigned mask_id)
{
	Subnet  mesh(0xff);
	
	mesh.Mask(mask_id);
	vnet_virtualize(vnet, s.getBuffer(), mesh.getBuffer(), 0);
}
*/
void Vnet::Virtualize(VSubnet& vs)
{
	vnet_virtualize_merge(vnet, vs.getBuffer(), vnet);
	// (re)initialize internal context stack
	if (vnet->context_stack != NULL)
		vc_free(vnet->context_stack);
	vnet->context_stack = newVContextStack(vnet);
}

Vnet& Vnet::operator=(const Vnet& v)
{
    vnet_copy(vnet, v.vnet);
	// initialize internal context stack
	if (vnet->context_stack != NULL)
		vc_free(vnet->context_stack);
	vnet->context_stack = newVContextStack(vnet);
    return *this;
}

void Vnet::Print(unsigned flag)
{
    vnet_print(vnet, Width(), flag);
}

/* Set context status */
void Vnet::Everywhere(const Vnet& v)
{
	if ((v.vnet)->context_stack->current == NULL) {
		fprintf(stderr, "VNET> Warning : No context stack, EVERYWHERE not processed\n"); 
	} else
		VCONTEXT_SET_EVERYWHERE(((v.vnet)->context_stack->current), v.ComponentCount());
}

void Vnet::Where(const VPvarBool& cond, const Vnet& v)
{
	if ((v.vnet)->context_stack->current == NULL) {
		fprintf(stderr, "VNET> Warning : No context stack, EVERYWHERE not processed\n"); 
	} else
		VCONTEXT_WHERE(((v.vnet)->context_stack), cond.getBuffer());
}

void Vnet::Endwhere(const Vnet& v)
{
	if ((v.vnet)->context_stack->current == NULL) {
		fprintf(stderr, "VNET> Warning : No context stack, EVERYWHERE not processed\n"); 
	} else
		VCONTEXT_ENDWHERE(((v.vnet)->context_stack));
}

void Vnet::Elsewhere(const Vnet& v)
{
	if ((v.vnet)->context_stack->current == NULL) {
		fprintf(stderr, "VNET> Warning : No context stack, EVERYWHERE not processed\n"); 
	} else
		VCONTEXT_ELSEWHERE(((v.vnet)->context_stack));
}

/* Set context status */
void Vnet::Everywhere()
{
	if (Vnet::Implicit() == NULL) {
		fprintf(stderr, "VNET> Fatal Error : no implicit virtual net declared. Program terminated\n");
		exit(1);
	} else
		Vnet::Everywhere(*(Vnet::Implicit()));
}
void Vnet::Where(const VPvarBool& cond) {
	if (Vnet::Implicit() == NULL) {
		fprintf(stderr, "VNET> Fatal Error : no implicit virtual net declared. Program terminated\n");
		exit(1);
	} else
		Vnet::Where(cond, *(Vnet::Implicit()));
}

void Vnet::Endwhere() {
	if (Vnet::Implicit() == NULL) {
		fprintf(stderr, "VNET> Fatal Error : no implicit virtual net declared. Program terminated\n");
		exit(1);
	} else
		Vnet::Endwhere(*(Vnet::Implicit()));
}

void Vnet::Elsewhere() {
	if (Vnet::Implicit() == NULL) {
		fprintf(stderr, "VNET> Fatal Error : no implicit virtual net declared. Program terminated\n");
		exit(1);
	} else
		Vnet::Elsewhere(*(Vnet::Implicit()));
}

